#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2024, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

"""Ansible module to operate the Authentication and Policy Servers in Cisco Catalyst Center."""
from __future__ import absolute_import, division, print_function

__metaclass__ = type
__author__ = ["Muthu Rakesh, Madhan Sankaranarayanan"]

DOCUMENTATION = r"""
---
module: ise_radius_integration_workflow_manager
short_description: Resource module for Authentication and Policy Servers
description:
- Manage operations on Authentication and Policy Servers.
- API to create Authentication and Policy Server Access Configuration.
- API to update Authentication and Policy Server Access Configuration.
- API to delete Authentication and Policy Server Access Configuration.
version_added: '6.14.0'
extends_documentation_fragment:
  - cisco.dnac.workflow_manager_params
author: Muthu Rakesh (@MUTHU-RAKESH-27)
        Madhan Sankaranarayanan (@madhansansel)
options:
  config_verify:
    description: Set to True to verify the Cisco Catalyst Center after applying the playbook config.
    type: bool
    default: False
  state:
    description: The state of Cisco Catalyst Center after module completion.
    type: str
    choices: [ "merged", "deleted" ]
    default: merged
  config:
    description:
    - List of details of Authentication and Policy Servers being managed.
    type: list
    elements: dict
    required: true
    suboptions:
      authentication_policy_server:
        description: Manages the Authentication and Policy Servers.
        type: list
        elements: dict
        suboptions:
          server_type:
            description:
            - Type of the Authentication and Policy Server.
            - ISE for Cisco ISE servers.
            - AAA for Non-Cisco ISE servers.
            type: str
            choices: [ "AAA", "ISE" ]
            default: AAA
          server_ip_address:
            description: IP Address of the Authentication and Policy Server.
            type: str
            required: True
          shared_secret:
            description:
            - Shared secret between devices and authentication and policy server.
            - Shared secret must have 4 to 100 characters with no spaces or the following characters - ["<", "?"].
            - Shared secret is a Read-Only parameter.
            type: str
          protocol:
            description:
            - Type of protocol for authentication and policy server.
            - RADIUS provides centralized services (AAA) for users in remote access scenarios.
            - TACACS focuses on access control and administrative authentication for network devices.
            type: str
            choices: [ "TACACS", "RADIUS", "RADIUS_TACACS" ]
            default: RADIUS
          encryption_scheme:
            description:
            - Type of encryption scheme for additional security.
            - If encryption scheme is given, then message authenticator code and encryption keys need to be required.
            - Updation of encryption scheme is not possible.
            - >
              KEYWRAP is used for securely wrapping and unwrapping encryption keys,
              ensuring their confidentiality during transmission or storage.
            - >
              RADSEC is an extension of RADIUS that provides secure communication
              between RADIUS clients and servers over TLS/SSL. Enhances enhancing the
              confidentiality and integrity of authentication and accounting data exchange.
            type: str
            choices: [ "KEYWRAP", "RADSEC" ]
          encryption_key:
            description:
            - Encryption key used to encrypt shared secret.
            - Updation of encryption scheme is not possible.
            - Required when encryption_scheme is provided.
            - >
              When ASCII format is selected, Encryption Key may contain
              alphanumeric and special characters. Key must be 16 char long.
            type: str
          message_authenticator_code_key:
            description:
            - Message key used to encrypt shared secret.
            - Updation of message key is not possible.
            - Required when encryption_scheme is provided.
            - >
              Message Authentication Code Key may contain alphanumeric and special characters.
              Key must be 20 char long.
            type: str
          authentication_port:
            description:
            - Authentication port of RADIUS server.
            - Updation of authentication port is not possible.
            - Authentication port should be from 1 to 65535.
            type: int
            default: 1812
          accounting_port:
            description:
            - Accounting port of RADIUS server.
            - Updation of accounting port is not possible.
            - Accounting port should be from 1 to 65535.
            type: int
            default: 1813
          retries:
            description:
            - Number of communication retries between devices and authentication and policy server.
            - Retries should be from 1 to 3.
            type: int
            default: 3
          timeout:
            description:
            - Number of seconds before timing out between devices and authentication and policy server.
            - Timeout should be from 2 to 20.
            type: int
            default: 4
          role:
            description:
            - Role of authentication and policy server.
            - Updation of role is not possible
            type: str
            default: secondary
          pxgrid_enabled:
            description:
            - Set True to enable the Pxgrid and False to disable the Pxgrid.
            - Pxgrid is available only for the Cisco ISE Servers.
            - >
              PxGrid facilitates seamless integration and information sharing across products,
              enhancing threat detection and response capabilities within the network ecosystem.
            type: bool
            default: True
          use_dnac_cert_for_pxgrid:
            description: Set True to use the Cisco Catalyst Center certificate for the Pxgrid.
            type: bool
            default: False
          cisco_ise_dtos:
            description:
            - List of Cisco ISE Data Transfer Objects (DTOs).
            - Required when server_type is set to ISE.
            type: list
            elements: dict
            suboptions:
              user_name:
                description:
                - User name of the Cisco ISE server.
                - Required for passing the cisco_ise_dtos.
                type: str
              password:
                description:
                - Password of the Cisco ISE server.
                - Password must have 4 to 127 characters with no spaces or the following characters - "<".
                - Required for passing the cisco_ise_dtos.
                type: str
              fqdn:
                description:
                - Fully-qualified domain name of the Cisco ISE server.
                - Required for passing the cisco_ise_dtos.
                type: str
              ip_address:
                description:
                - IP Address of the Cisco ISE Server.
                - Required for passing the cisco_ise_dtos.
                type: str
              description:
                description: Description about the Cisco ISE server.
                type: str
              ssh_key:
                description: SSH key of the Cisco ISE server.
                type: str
          external_cisco_ise_ip_addr_dtos:
            description: External Cisco ISE IP address data transfer objects for future use.
            type: list
            elements: dict
            suboptions:
              external_cisco_ise_ip_addresses:
                description: External Cisco ISE IP addresses.
                type: list
                elements: dict
                suboptions:
                  external_ip_address:
                    description: External Cisco ISE IP address.
                    type: str
              ise_type:
                description: Type of the Authentication and Policy Server.
                type: str
          trusted_server:
            description:
            - Indicates whether the certificate is trustworthy for the server.
            - Serves as a validation of its authenticity and reliability in secure connections.
            default: True
            type: bool
          ise_integration_wait_time:
            description:
            - Indicates the sleep time after initiating the Cisco ISE integration process.
            - Maximum sleep time should be less or equal to 120 seconds.
            default: 20
            type: int
requirements:
- dnacentersdk >= 2.7.2
- python >= 3.9
notes:
  - SDK Method used are
    system_settings.SystemSettings.add_authentication_and_policy_server_access_configuration,
    system_settings.SystemSettings.edit_authentication_and_policy_server_access_configuration,
    system_settings.SystemSettings.accept_cisco_ise_server_certificate_for_cisco_ise_server_integration,
    system_settings.SystemSettings.delete_authentication_and_policy_server_access_configuration,
    system_settings.SystemSettings.get_authentication_and_policy_servers,
    system_settings.SystemSettings.cisco_ise_server_integration_status,

  - Paths used are
    post /dna/intent/api/v1/authentication-policy-servers,
    put /dna/intent/api/v1/authentication-policy-servers/${id},
    put /dna/intent/api/v1/integrate-ise/${id},
    delete /dna/intent/api/v1/authentication-policy-servers/${id}
    get /dna/intent/api/v1/authentication-policy-servers
    get /dna/intent/api/v1/ise-integration-status

"""

EXAMPLES = r"""
- name: Create an AAA server.
  cisco.dnac.ise_radius_integration_workflow_manager:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    dnac_log: True
    dnac_log_level: "{{ dnac_log_level }}"
    state: merged
    config_verify: True
    config:
    - authentication_policy_server:
      - server_type: AAA
        server_ip_address: 10.0.0.1
        shared_secret: "12345"
        protocol: RADIUS_TACACS
        encryption_scheme: KEYWRAP
        encryption_key: "1234567890123456"
        message_authenticator_code_key: asdfghjklasdfghjklas
        authentication_port: 1812
        accounting_port: 1813
        retries: 3
        timeout: 4
        role: secondary

- name: Create an Cisco ISE server.
  cisco.dnac.ise_radius_integration_workflow_manager:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    dnac_log: True
    dnac_log_level: "{{ dnac_log_level }}"
    state: merged
    config_verify: True
    config:
    - authentication_policy_server:
      - server_type: ISE
        server_ip_address: 10.0.0.2
        shared_secret: "12345"
        protocol: RADIUS_TACACS
        encryption_scheme: KEYWRAP
        encryption_key: "1234567890123456"
        message_authenticator_code_key: asdfghjklasdfghjklas
        authentication_port: 1812
        accounting_port: 1813
        retries: 3
        timeout: 4
        role: primary
        use_dnac_cert_for_pxgrid: False
        pxgrid_enabled: True
        cisco_ise_dtos:
        - user_name: Cisco ISE
          password: "12345"
          fqdn: abs.cisco.com
          ip_address: 10.0.0.2
          description: Cisco ISE
        trusted_server: True
        ise_integration_wait_time: 20

- name: Update an AAA server.
  cisco.dnac.ise_radius_integration_workflow_manager:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    dnac_log: True
    dnac_log_level: "{{ dnac_log_level }}"
    state: merged
    config_verify: True
    config:
    - authentication_policy_server:
      - server_type: AAA
        server_ip_address: 10.0.0.1
        protocol: RADIUS_TACACS
        retries: 3
        timeout: 5

- name: Update an Cisco ISE server.
  cisco.dnac.ise_radius_integration_workflow_manager:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    dnac_log: True
    dnac_log_level: "{{ dnac_log_level }}"
    state: merged
    config_verify: True
    config:
    - authentication_policy_server:
      - server_type: ISE
        server_ip_address: 10.0.0.2
        protocol: RADIUS_TACACS
        retries: 3
        timeout: 5
        use_dnac_cert_for_pxgrid: False
        pxgrid_enabled: True
        cisco_ise_dtos:
        - user_name: Cisco ISE
          password: "12345"
          fqdn: abs.cisco.com
          ip_address: 10.0.0.2
          description: Cisco ISE

- name: Delete an Authentication and Policy server.
  cisco.dnac.ise_radius_integration_workflow_manager:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    dnac_log: True
    dnac_log_level: "{{ dnac_log_level }}"
    state: deleted
    config_verify: True
    config:
    - authentication_policy_server:
      - server_ip_address: 10.0.0.1
"""

RETURN = r"""
# Case_1: Successful creation of Authentication and Policy Server.
response_1:
  description: A dictionary or list with the response returned by the Cisco Catalyst Center Python SDK
  returned: always
  type: dict
  sample: >
    {
      "response": {
        "taskId": "string",
        "url": "string"
      },
      "version": "string"
    }

# Case_2: Successful updation of Authentication and Policy Server.
response_2:
  description: A dictionary or list with the response returned by the Cisco Catalyst Center Python SDK
  returned: always
  type: dict
  sample: >
    {
      "response": {
        "taskId": "string",
        "url": "string"
      },
      "version": "string"
    }

# Case_3: Successful creation/updation of network
response_3:
  description: A dictionary or list with the response returned by the Cisco Catalyst Center Python SDK
  returned: always
  type: dict
  sample: >
    {
      "response": {
        "taskId": "string",
        "url": "string"
      },
      "version": "string"
    }
"""

import copy
import time
from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.dnac.plugins.module_utils.dnac import (
    DnacBase,
    validate_list_of_dicts,
    get_dict_result,
    dnac_compare_equality,
)


class IseRadiusIntegration(DnacBase):
    """Class containing member attributes for ise_radius_integration_workflow_manager module"""

    def __init__(self, module):
        super().__init__(module)
        self.max_timeout = self.params.get('dnac_api_task_timeout')
        self.result["response"] = [
            {"authenticationPolicyServer": {"response": {}, "msg": {}}}
        ]
        self.authentication_policy_server_obj_params = \
            self.get_obj_params("authenticationPolicyServer")
        self.validation_string = ""

    def validate_input(self):
        """
        Checks if the configuration parameters provided in the playbook
        meet the expected structure and data types,
        as defined in the 'temp_spec' dictionary.

        Parameters:
            None

        Returns:
            self

        """

        if not self.config:
            self.msg = "config not available in playbook for validation"
            self.status = "success"
            return self

        # temp_spec is the specification for the expected structure of configuration parameters
        temp_spec = {
            "authentication_policy_server": {
                "type": "list",
                "elements": "dict",
                "server_type": {"type": 'string', "choices": ["AAA", "ISE"]},
                "server_ip_address": {"type": 'string'},
                "shared_secret": {"type": 'string'},
                "protocol": {"type": 'string', "choices": ["TACACS", "RADIUS", "RADIUS_TACACS"]},
                "encryption_scheme": {"type": 'string'},
                "message_authenticator_code_key": {"type": 'string'},
                "encryption_key": {"type": 'string'},
                "authentication_port": {"type": 'integer'},
                "accounting_port": {"type": 'integer'},
                "retries": {"type": 'integer'},
                "timeout": {"type": 'integer'},
                "role": {"type": 'string'},
                "pxgrid_enabled": {"type": 'bool'},
                "use_dnac_cert_for_pxgrid": {"type": 'bool'},
                "cisco_ise_dtos": {
                    "type": 'list',
                    "user_name": {"type": 'string'},
                    "password": {"type": 'string'},
                    "fqdn": {"type": 'string'},
                    "ip_address": {"type": 'string'},
                    "description": {"type": 'string'},
                    "ssh_key": {"type": 'string'},
                },
                "external_cisco_ise_ip_addr_dtos": {
                    "type": 'list',
                    "external_cisco_ise_ip_addresses": {
                        "type": 'list',
                        "external_ip_address": {"type": 'string'},
                    },
                    "ise_type": {"type": 'string'},
                },
                "trusted_server": {"type": 'bool'},
                "ise_integration_wait_time": {"type": 'integer'}
            }
        }

        # Validate playbook params against the specification (temp_spec)
        valid_temp, invalid_params = validate_list_of_dicts(self.config, temp_spec)
        if invalid_params:
            self.msg = "Invalid parameters in playbook: {0}".format("\n".join(invalid_params))
            self.status = "failed"
            return self

        self.validated_config = valid_temp
        self.log("Successfully validated playbook config params: {0}".format(valid_temp), "INFO")
        self.msg = "Successfully validated input from the playbook"
        self.status = "success"
        return self

    def requires_update(self, have, want, obj_params):
        """
        Check if the template config given requires update by comparing
        current information wih the requested information.

        This method compares the current global pool, reserve pool,
        or network details from Cisco Catalyst Center with the user-provided details
        from the playbook, using a specified schema for comparison.

        Parameters:
            have (dict) - Current information from the Cisco Catalyst Center
                          (global pool, reserve pool, network details)
            want (dict) - Users provided information from the playbook
            obj_params (list of tuples) - A list of parameter mappings specifying which
                                          Cisco Catalyst Center parameters (dnac_param)
                                          correspond to the user-provided
                                          parameters (ansible_param).

        Returns:
            bool - True if any parameter specified in obj_params differs between
            current_obj and requested_obj, indicating that an update is required.
            False if all specified parameters are equal.

        """

        current_obj = have
        requested_obj = want
        self.log("Current State (have): {0}".format(current_obj), "DEBUG")
        self.log("Desired State (want): {0}".format(requested_obj), "DEBUG")

        return any(not dnac_compare_equality(current_obj.get(dnac_param),
                                             requested_obj.get(ansible_param))
                   for (dnac_param, ansible_param) in obj_params)

    def get_obj_params(self, get_object):
        """
        Get the required comparison obj_params value

        Parameters:
            get_object (str) - identifier for the required obj_params

        Returns:
            obj_params (list) - obj_params value for comparison.
        """

        try:
            obj_params = []
            if get_object == "authenticationPolicyServer":
                obj_params = [
                    ("protocol", "protocol"),
                    ("retries", "retries"),
                    ("timeoutSeconds", "timeoutSeconds"),
                    ("pxgridEnabled", "pxgridEnabled"),
                    ("useDnacCertForPxgrid", "useDnacCertForPxgrid"),
                ]
            else:
                raise ValueError("Received an unexpected value for 'get_object': {0}"
                                 .format(get_object))
        except Exception as msg:
            self.log("Received exception: {0}".format(msg), "CRITICAL")

        return obj_params

    def get_auth_server_params(self, auth_server_info):
        """
        Process Authentication and Policy Server params from playbook data for
        Authentication and Policy Server config in Cisco Catalyst Center.

        Parameters:
            auth_server_info (dict) - Cisco Catalyst Center data containing
            information about the Authentication and Policy Server.

        Returns:
            dict or None - Processed Authentication and Policy Server data in a format suitable
            for Cisco Catalyst Center configuration, or None if auth_server_info is empty.
        """

        if not auth_server_info:
            self.log("Authentication and Policy Server data is empty", "INFO")
            return None

        self.log("Authentication and Policy Server Details: {0}".format(auth_server_info), "DEBUG")
        auth_server = {
            "authenticationPort": auth_server_info.get("authenticationPort"),
            "accountingPort": auth_server_info.get("accountingPort"),
            "isIseEnabled": auth_server_info.get("iseEnabled"),
            "ipAddress": auth_server_info.get("ipAddress"),
            "pxgridEnabled": auth_server_info.get("pxgridEnabled"),
            "useDnacCertForPxgrid": auth_server_info.get("useDnacCertForPxgrid"),
            "port": auth_server_info.get("port"),
            "protocol": auth_server_info.get("protocol"),
            "retries": str(auth_server_info.get("retries")),
            "role": auth_server_info.get("role"),
            "timeoutSeconds": str(auth_server_info.get("timeoutSeconds")),
            "encryptionScheme": auth_server_info.get("encryptionScheme"),
            "state": auth_server_info.get("state")
        }
        self.log("Formated Authentication and Policy Server details: {0}"
                 .format(auth_server), "DEBUG")
        if auth_server.get("isIseEnabled") is True:
            auth_server_ise_info = auth_server_info.get("ciscoIseDtos")
            auth_server.update({"ciscoIseDtos": []})
            for ise_credential in auth_server_ise_info:
                auth_server.get("ciscoIseDtos").append({
                    "userName": ise_credential.get("userName"),
                    "fqdn": ise_credential.get("fqdn"),
                    "ipAddress": ise_credential.get("ipAddress"),
                    "subscriberName": ise_credential.get("subscriberName"),
                    "description": ise_credential.get("description")
                })

        return auth_server

    def auth_server_exists(self, ipAddress):
        """
        Check if the Authentication and Policy Server with the given ipAddress exists

        Parameters:
            ipAddress (str) - The ipAddress of the Authentication and
                              Policy Server to check for existence.

        Returns:
            dict - A dictionary containing information about the
                   Authentication and Policy Server's existence:
            - 'exists' (bool): True if the Authentication and Policy Server exists, False otherwise.
            - 'id' (str or None): The ID of the Authentication and Policy Server if it exists
                                  or None if it doesn't.
            - 'details' (dict or None): Details of the Authentication and Policy Server if it exists
                                        else None.
        """

        AuthServer = {
            "exists": False,
            "details": None,
            "id": None
        }
        response = self.dnac._exec(
            family="system_settings",
            function='get_authentication_and_policy_servers',
        )
        if not isinstance(response, dict):
            self.log("Failed to retrieve the Authentication and Policy Server details - "
                     "Response is not a dictionary", "CRITICAL")
            return AuthServer

        all_auth_server_details = response.get("response")
        auth_server_details = get_dict_result(all_auth_server_details, "ipAddress", ipAddress)
        self.log("Authentication and Policy Server Ip Address: {0}"
                 .format(ipAddress), "DEBUG")
        self.log("Authentication and Policy Server details: {0}"
                 .format(auth_server_details), "DEBUG")
        if not auth_server_details:
            self.log("Authentication and Policy Server {0} does not exist".format(ipAddress), "INFO")
            return AuthServer

        AuthServer.update({"exists": True})
        AuthServer.update({"id": auth_server_details.get("instanceUuid")})
        AuthServer["details"] = self.get_auth_server_params(auth_server_details)

        self.log("Formatted Authenticaion and Policy Server details: {0}".format(AuthServer), "DEBUG")
        return AuthServer

    def get_have_authentication_policy_server(self, authentication_policy_server):
        """
        Get the current Authentication and Policy Server information from
        Cisco Catalyst Center based on the provided playbook details.
        check this API using check_return_status.

        Parameters:
            authentication_policy_server (list of dict) - Playbook details containing
            Authentication and Policy Server configuration.

        Returns:
            self - The current object with updated
            Authentication and Policy Server information.
        """

        all_auth_server_details = []
        for item in authentication_policy_server:
            auth_server = {
                "exists": False,
                "details": None,
                "id": None
            }
            ip_address = item.get("server_ip_address")
            if ip_address is None:
                self.msg = "The parameter 'server_ip_address' is missing but required."
                self.status = "failed"
                return self

            auth_server = self.auth_server_exists(ip_address)
            self.log("Authentication and Policy Server exists for '{0}': {1}"
                     .format(ip_address, auth_server.get("exists")), "DEBUG")
            self.log("Authentication and Policy Server details for '{0}': {1}"
                     .format(ip_address, auth_server.get("details")), "DEBUG")
            self.log("Authentication and Policy Server Id for '{0}': {1}"
                     .format(ip_address, auth_server.get("id")), "DEBUG")
            all_auth_server_details.append(auth_server)

        self.have.update({"authenticationPolicyServer": all_auth_server_details})
        self.msg = "Collecting the Authentication and Policy Server " + \
                   "details from the Cisco Catalyst Center."
        self.status = "success"
        return self

    def get_have(self, config):
        """
        Get the current Authentication and Policy Server details from Cisco Catalyst Center

        Parameters:
            config (dict) - Playbook details containing
            Authentication and Policy Server configuration.

        Returns:
            self - The current object with updated
            Authentication and Policy Server information.
        """

        authentication_policy_server = config.get("authentication_policy_server")
        if authentication_policy_server is None:
            self.msg = "The 'authentication_policy_server' is missing in the playbook configuration."
            self.status = "failed"
            return self

        self.get_have_authentication_policy_server(authentication_policy_server).check_return_status()
        self.log("Current State (have): {0}".format(self.have), "INFO")
        self.msg = "Successfully retrieved the details from the Cisco Catalyst Center"
        self.status = "success"
        return self

    def get_primary_ise_fqdn(self, cisco_ise_dtos, ip_address):
        """
        Get the fully qualified domain name (fqdn) from the Multi-Node ISE environment
        in the Cisco Catalyst Center with the role as 'primary'.

        Parameters:
            cisco_ise_dtos (list of dict): List of configuration of Multi-Node in an
            ISE environment.
            ip_address (str): IP address of the Cisco ISE server.

        Returns:
            primary_fqdn (str): The fqdn value of the primary node of the ISE server.
        """

        self.log(
            "Getting the FQDN (Fully Qualified Domain Name) of the primary node of the ISE server. "
            "Input parameters - cisco_ise_dtos: {cisco_ise_dtos}, ip_address: {ip_address}"
            .format(cisco_ise_dtos=cisco_ise_dtos, ip_address=ip_address)
        )
        primary_node_ise_details = get_dict_result(cisco_ise_dtos, "role", "PRIMARY")
        if not primary_node_ise_details:
            self.msg = (
                "There are no details of the 'PRIMARY' role node in the multi-node ISE environment {ip_address}."
                .format(ip_address=ip_address)
            )
            self.log(str(self.msg), "ERROR")
            self.status = "failed"
            return self.check_return_status()

        primary_fqdn = primary_node_ise_details.get("fqdn")
        if not primary_fqdn:
            self.msg = (
                "The FQDN value is not available in the 'primary' node with IP address'{ip_address}'."
                .format(ip_address=ip_address)
            )
            self.log(str(self.msg), "ERROR")
            self.status = "failed"
            return self.check_return_status()

        self.log(
            "Returning the FQDN value of the primary node '{ip_address}': {primary_fqdn}"
            .format(ip_address=ip_address, primary_fqdn=primary_fqdn)
        )
        return primary_fqdn

    def get_want_authentication_policy_server(self, auth_policy_server):
        """
        Get all the Authentication Policy Server information from playbook
        Set the status and the msg before returning from the API
        Check the return value of the API with check_return_status()

        Parameters:
            auth_policy_server (list of dict) - Playbook authentication policy server details
            containing IpAddress, authentication port, accounting port, Cisco ISE Details,
            protocol, retries, role, timeout seconds, encryption details.

        Returns:
            self - The current object with updated desired Authentication Policy Server information.
        """

        all_auth_server_details = []
        auth_server_index = 0
        for item in auth_policy_server:
            auth_server = {}
            auth_server_exists = self.have.get("authenticationPolicyServer")[auth_server_index].get("exists")
            auth_server_details = self.have.get("authenticationPolicyServer")[auth_server_index].get("details")
            trusted_server = False
            if not auth_server_exists:
                server_type = item.get("server_type")
                if server_type not in ["ISE", "AAA", None]:
                    self.msg = "The 'server_type' should be either 'ISE' or 'AAA' but {0} was provided.".format(server_type)
                    self.status = "failed"
                    return self

                if server_type == "ISE":
                    auth_server.update({"isIseEnabled": True})
                else:
                    auth_server.update({"isIseEnabled": False})
            else:
                auth_server.update({"isIseEnabled": auth_server_details.get("isIseEnabled")})

            auth_server.update({"ipAddress": item.get("server_ip_address")})

            auth_server_exists = self.have.get("authenticationPolicyServer")[auth_server_index].get("exists")

            if not auth_server_exists:
                shared_secret = item.get("shared_secret")
                if not shared_secret:
                    self.msg = "The required parameter 'shared_secret' is missing."
                    self.status = "failed"
                    return self

                shared_secret = str(shared_secret)
                if len(shared_secret) < 4 or len(shared_secret) > 100:
                    self.msg = "The 'shared_secret' should contain between 4 and 100 characters."
                    self.status = "failed"
                    return self

                invalid_chars = " ?<"
                for char in invalid_chars:
                    if char in shared_secret:
                        self.msg = "The 'shared_secret' should not contain spaces or the characters '?', '<'."
                        self.status = "failed"
                        return self

                auth_server.update({"sharedSecret": shared_secret})

            protocol = item.get("protocol")
            if protocol not in ["RADIUS", "TACACS", "RADIUS_TACACS", None]:
                self.msg = (
                    "The 'protocol' should be one of ['RADIUS', 'TACACS', 'RADIUS_TACACS'], "
                    "not '{0}'.".format(protocol)
                )
                self.status = "failed"
                return self

            if protocol is not None:
                auth_server.update({"protocol": protocol})
            else:
                if not auth_server_exists:
                    auth_server.update({"protocol": "RADIUS"})
                else:
                    auth_server.update({"protocol": auth_server_details.get("protocol")})

            auth_server.update({"port": 49})

            if not auth_server_exists:
                encryption_scheme = item.get("encryption_scheme")
                if encryption_scheme not in ["KEYWRAP", "RADSEC", None]:
                    self.msg = "The 'encryption_scheme' should be in ['KEYWRAP', 'RADSEC']. " + \
                               "It should not be {0}.".format(encryption_scheme)
                    self.status = "failed"
                    return self

                if encryption_scheme:
                    auth_server.update({"encryptionScheme": encryption_scheme})

                if encryption_scheme == "KEYWRAP":
                    message_key = item.get("message_authenticator_code_key")
                    if not message_key:
                        self.msg = "The 'message_authenticator_code_key' must not be empty when the 'encryption_scheme' is set to 'KEYWRAP'."
                        self.status = "failed"
                        return self

                    message_key = str(message_key)
                    message_key_length = len(message_key)
                    if message_key_length != 20:
                        self.msg = "The 'message_authenticator_code_key' should be exactly 20 characters."
                        self.status = "failed"
                        return self

                    auth_server.update({"messageKey": message_key})

                    encryption_key = item.get("encryption_key")
                    if not encryption_key:
                        self.msg = "The 'encryption_key' should not be empty if encryption_scheme is 'KEYWRAP'."
                        self.status = "failed"
                        return self

                    encryption_key = str(encryption_key)
                    encryption_key_length = len(encryption_key)
                    if encryption_key_length != 16:
                        self.msg = "The 'encryption_key' must be 16 characters long. It may contain alphanumeric and special characters."
                        self.status = "failed"
                        return self

                    auth_server.update({"encryptionKey": encryption_key})

            if not auth_server_exists:
                authentication_port = item.get("authentication_port")
                if not authentication_port:
                    authentication_port = 1812

                if not str(authentication_port).isdigit():
                    self.msg = "The 'authentication_port' should contain only digits."
                    self.status = "failed"
                    return self

                if authentication_port < 1 or authentication_port > 65535:
                    self.msg = "The 'authentication_port' should be from 1 to 65535."
                    self.status = "failed"
                    return self

                auth_server.update({"authenticationPort": authentication_port})
            else:
                auth_server.update({"authenticationPort": auth_server_details.get("authenticationPort")})

            if not auth_server_exists:
                accounting_port = item.get("accounting_port")
                if not accounting_port:
                    accounting_port = 1813

                if not str(accounting_port).isdigit():
                    self.msg = "The 'accounting_port' should contain only digits."
                    self.status = "failed"
                    return self

                if accounting_port < 1 or accounting_port > 65535:
                    self.msg = "The 'accounting_port' should be from 1 to 65535."
                    self.status = "failed"
                    return self

                auth_server.update({"accountingPort": accounting_port})
            else:
                auth_server.update({"accountingPort": auth_server_details.get("accountingPort")})

            retries = item.get("retries")
            if not retries:
                if not auth_server_exists:
                    auth_server.update({"retries": "3"})
                else:
                    auth_server.update({"retries": auth_server_details.get("retries")})
            else:
                try:
                    retries_int = int(retries)
                    if retries_int < 1 or retries_int > 3:
                        self.msg = "The 'retries' should be from 1 to 3."
                        self.status = "failed"
                        return self
                except ValueError:
                    self.msg = "The 'retries' should contain only from 0-9."
                    self.status = "failed"
                    return self

                auth_server.update({"retries": str(retries)})

            timeout = item.get("timeout")
            if not auth_server_exists:
                default_timeout = "4"
            else:
                default_timeout = str(auth_server_details.get("timeoutSeconds"))

            # If 'timeout' is not provided, use 'default_timeout'
            if timeout is None:
                auth_server.update({"timeoutSeconds": default_timeout})
            else:
                try:
                    timeout_int = int(timeout)
                    if timeout_int < 2 or timeout_int > 20:
                        self.msg = "The 'timeout' should be from 2 to 20."
                        self.status = "failed"
                        return self

                    auth_server.update({"timeoutSeconds": str(timeout)})
                except ValueError:
                    self.msg = "The 'time_out' must contain only digits."
                    self.status = "failed"
                    return self

            # Determine the role based on whether the auth server exists and if the role is specified
            if not auth_server_exists:
                # Use the role from 'auth_policy_server' if available, otherwise default to "secondary"
                role = item.get("role", "secondary")
            else:
                # Use the role from 'auth_server_details'
                role = auth_server_details.get("role")

            auth_server.update({"role": role})

            if auth_server.get("isIseEnabled"):
                cisco_ise_dtos = item.get("cisco_ise_dtos")
                if not cisco_ise_dtos:
                    self.msg = "The required parameter 'cisco_ise_dtos' is missing for 'server_type' is 'ISE'."
                    self.status = "failed"
                    return self

                auth_server.update({"ciscoIseDtos": []})
                position_ise_creds = 0
                for ise_credential in cisco_ise_dtos:
                    auth_server.get("ciscoIseDtos").append({})
                    user_name = ise_credential.get("user_name")
                    if not user_name:
                        if not auth_server_exists:
                            self.msg = "The required parameter 'user_name' is missing when 'server_type' is 'ISE'."
                            self.status = "failed"
                            return self

                        user_name = auth_server_details.get("ciscoIseDtos")[0].get("userName")

                    auth_server.get("ciscoIseDtos")[position_ise_creds].update({
                        "userName": user_name
                    })

                    password = ise_credential.get("password")
                    if not password:
                        self.msg = "The required parameter 'password' is missing when 'server_type' is 'ISE'."
                        self.status = "failed"
                        return self

                    if not 4 <= len(password) <= 127:
                        self.msg = ""
                        self.status = "failed"
                        return self

                    auth_server.get("ciscoIseDtos")[position_ise_creds].update({
                        "password": password
                    })

                    ip_address = ise_credential.get("ip_address")
                    if not ip_address:
                        self.msg = "The required parameter 'ip_address' is missing when 'server_type' is 'ISE'."
                        self.status = "failed"
                        return self

                    auth_server.get("ciscoIseDtos")[position_ise_creds].update({
                        "ipAddress": ip_address
                    })

                    fqdn = ise_credential.get("fqdn")
                    if not fqdn:
                        self.log("FQDN is missing in the provided ISE credentials, proceeding to fetch it...", "DEBUG")
                        if not auth_server_exists:
                            self.msg = "The required parameter 'fqdn' is missing when 'server_type' is 'ISE'."
                            self.log(str(self.msg), "ERROR")
                            self.status = "failed"
                            return self

                        self.log(
                            "Auth server exists; retrieving authentication and policy servers to obtain FQDN for IP '{0}'."
                            .format(ip_address),
                            "DEBUG"
                        )
                        response = self.dnac._exec(
                            family="system_settings",
                            function='get_authentication_and_policy_servers',
                            params={"is_ise_enabled": True}
                        )
                        self.log("Successfully retrieved authentication server response; processing Cisco ISE details.", "DEBUG")
                        response = response.get("response")
                        if response is None:
                            self.msg = (
                                "Failed to retrieve information from 'get_authentication_and_policy_servers' for IP '{0}'."
                                .format(ip_address)
                            )
                            self.log(str(self.msg), "ERROR")
                            self.status = "failed"
                            return self

                        cisco_ise_dtos = response[0].get("ciscoIseDtos")
                        if not cisco_ise_dtos:
                            self.msg = "No Cisco ISE details available for IP '{0}'.".format(ip_address)
                            self.status = "failed"
                            self.log(str(self.msg), "ERROR")
                            return self

                        self.log("Cisco ISE details found; retrieving FQDN of primary node...", "DEBUG")
                        fqdn = self.get_primary_ise_fqdn(cisco_ise_dtos, ip_address)
                        self.log("Updating authentication server details with FQDN...", "DEBUG")
                        self.log(
                            "Retrieved FQDN for primary ISE node with IP '{ip}': {fqdn}"
                            .format(ip=ip_address, fqdn=fqdn),
                            "INFO"
                        )

                    auth_server.get("ciscoIseDtos")[position_ise_creds].update({"fqdn": fqdn})
                    self.log("Authentication server details successfully updated with FQDN.", "INFO")

                    if not auth_server_exists:
                        auth_server.get("ciscoIseDtos")[position_ise_creds].update({
                            "subscriberName": "ersadmin"
                        })
                    else:
                        auth_server.get("ciscoIseDtos")[position_ise_creds].update({
                            "subscriberName": auth_server_details.get("ciscoIseDtos")[0].get("subscriberName")
                        })

                    description = ise_credential.get("description")
                    if description:
                        auth_server.get("ciscoIseDtos")[position_ise_creds].update({
                            "description": description
                        })

                    ssh_key = ise_credential.get("ssh_key")
                    if ssh_key:
                        auth_server.get("ciscoIseDtos")[position_ise_creds].update({
                            "sshkey": str(ssh_key)
                        })

                    position_ise_creds += 1

                pxgrid_enabled = item.get("pxgrid_enabled")
                if pxgrid_enabled is None:
                    if auth_server_exists:
                        pxgrid_enabled = auth_server_details.get("pxgridEnabled")
                    else:
                        pxgrid_enabled = True

                auth_server.update({"pxgridEnabled": pxgrid_enabled})

                use_dnac_cert_for_pxgrid = item.get("use_dnac_cert_for_pxgrid")
                if use_dnac_cert_for_pxgrid is None:
                    if auth_server_exists:
                        use_dnac_cert_for_pxgrid = auth_server_details.get("useDnacCertForPxgrid")
                    else:
                        use_dnac_cert_for_pxgrid = False

                auth_server.update({"useDnacCertForPxgrid": use_dnac_cert_for_pxgrid})

                external_cisco_ise_ip_addr_dtos = item \
                    .get("external_cisco_ise_ip_addr_dtos")
                if external_cisco_ise_ip_addr_dtos:
                    auth_server.update({"externalCiscoIseIpAddrDtos": []})
                    position_ise_addresses = 0
                    for external_cisco_ise in external_cisco_ise_ip_addr_dtos:
                        external_cisco_ise_ip_addresses = external_cisco_ise \
                            .get("external_cisco_ise_ip_addresses")
                        if external_cisco_ise_ip_addresses:
                            auth_server.get("externalCiscoIseIpAddrDtos").append({})
                            auth_server.get("externalCiscoIseIpAddrDtos")[position_ise_addresses] \
                                .update({"externalCiscoIseIpAddresses": []})
                            position_ise_address = 0
                            for external_ip_address in external_cisco_ise_ip_addresses:
                                auth_server.get("externalCiscoIseIpAddrDtos")[position_ise_addresses] \
                                    .get("externalCiscoIseIpAddresses").append({})
                                auth_server.get("externalCiscoIseIpAddrDtos")[position_ise_addresses] \
                                    .get("externalCiscoIseIpAddresses")[position_ise_address].update({
                                        "externalIpAddress": external_ip_address.get("external_ip_address")
                                    })
                                position_ise_address += 1
                        ise_type = external_cisco_ise.get("ise_type")
                        if ise_type:
                            auth_server.get("externalCiscoIseIpAddrDtos")[position_ise_addresses] \
                                .update({"type": ise_type})
                        position_ise_addresses += 1

                trusted_server = item.get("trusted_server")
                if item.get("trusted_server") is None:
                    trusted_server = True
                else:
                    trusted_server = item.get("trusted_server")

                self.want.update({"trusted_server": trusted_server})

                ise_integration_wait_time = item.get("ise_integration_wait_time")
                if ise_integration_wait_time is None:
                    ise_integration_wait_time = 20
                else:
                    try:
                        ise_integration_wait_time_int = int(ise_integration_wait_time)
                        if ise_integration_wait_time_int < 1 or ise_integration_wait_time_int > 120:
                            self.msg = "The 'ise_integration_wait_time' should be from 1 to 120 seconds."
                            self.status = "failed"
                            return self

                    except ValueError:
                        self.msg = "The 'ise_integration_wait_time' should contain only digits."
                        self.status = "failed"
                        return self

                self.want.update({"ise_integration_wait_time": ise_integration_wait_time})

            self.log("Authentication and Policy Server playbook details: {0}"
                     .format(auth_server), "DEBUG")
            all_auth_server_details.append(auth_server)
            auth_server_index += 1

        self.want.update({"authenticationPolicyServer": all_auth_server_details})
        self.msg = "Collecting the Authentication and Policy Server details from the playbook"
        self.status = "success"
        return self

    def get_want(self, config):
        """
        Get all the Authentication Policy Server related information from playbook

        Parameters:
            config (list of dict) - Playbook details

        Returns:
            None
        """

        if config.get("authentication_policy_server"):
            auth_policy_server = config.get("authentication_policy_server")
            self.get_want_authentication_policy_server(auth_policy_server).check_return_status()

        self.log("Desired State (want): {0}".format(self.want), "INFO")
        self.msg = "Successfully retrieved details from the playbook"
        self.status = "success"
        return self

    def check_ise_server_integration_status(self, ip_address):
        """
        Check whether the Cisco ISE server is ready for the accepting the user authentication certificate.

        Parameters:
            ip_address (str) - The IP address of the Cisco ISE server.
            self - The current object with updated desired Authentication Policy Server information.

        Returns:
            self - The current object with updated desired Authentication Policy Server information.
        """

        start_time = time.time()
        while True:
            try:
                cisco_ise_status = self.dnac._exec(
                    family="system_settings",
                    function="cisco_ise_server_integration_status",
                    op_modifies=True,
                )
            except Exception as msg:
                self.msg = (
                    "Exception occurred while checking the status of the Cisco ISE server with IP address '{ip}'."
                    .format(ip=ip_address)
                )

            overall_status = cisco_ise_status.get("overallStatus")
            statuses = ["WAITING_USER_INPUT", "COMPLETE"]
            if overall_status in statuses:
                self.log("The status of the Cisco ISE server is '{status}'".format(status=overall_status))
                break

            if (time.time() - start_time) >= 10:
                self.msg = (
                    "The Cisco Catalyst Center took more than 10 seconds to accept "
                    "the PxGrid certificate of the Cisco ISE server with ."
                )
                self.status = "failed"
                break

            time.sleep(1)

        return self

    def accept_cisco_ise_server_certificate(self, ipAddress, trusted_server):
        """
        Accept the Cisco ISE server certificate in Cisco Catalyst
        Center provided in the playbook.

        Parameters:
            ipAddress (str) - The Ip address of the Authentication and Policy Server to be deleted.
            trusted_server (bool) - Indicates whether the certificate is trustworthy for the server.

        Returns:
            None
        """

        try:
            AuthServer = self.auth_server_exists(ipAddress)
            if not AuthServer:
                self.msg = "Error while retrieving the Authentication and Policy Server {0} \
                            details.".format(ipAddress)
                self.log(str(self.msg, "CRITICAL"))
                self.status = "failed"
                return self

            cisco_ise_id = AuthServer.get("id")
            if not cisco_ise_id:
                self.msg = "Error while retrieving the Authentication and Policy Server {0} id." \
                           .format(ipAddress)
                self.log(str(self.msg, "CRITICAL"))
                self.status = "failed"
                return self

            response = self.dnac._exec(
                family="system_settings",
                function="accept_cisco_ise_server_certificate_for_cisco_ise_server_integration",
                params={
                    "id": cisco_ise_id,
                    "isCertAcceptedByUser": trusted_server
                },
            )
            self.log("Received API response for 'accept_cisco_ise_server_certificate_"
                     "for_cisco_ise_server_integration': {0}".format(response), "DEBUG")
        except Exception as msg:
            self.log("Exception occurred while accepting the certificate of {0}: {1}"
                     .format(ipAddress, msg))
            return None

        return

    def check_auth_server_response_status(self, response, validation_string_set, api_name):
        """
        Checks the status of a task related to updating the authentication and policy server
        by polling the task details until it completes or a timeout is reached.

        Parameters:
            response (dict): The initial response from the task creation API.
            validation_string_set (set): A set of strings expected to be found in the task progress for a successful operation.
            api_name (str): Name of the function during the SDK call.

        Returns:
            self - The current object with updated desired Authentication Policy Server information.
        """

        response = response.get("response")
        if response.get("errorcode") is not None:
            self.msg = response.get("detail")
            self.status = "failed"
            return self

        task_id = response.get("taskId")
        start_time = time.time()
        sleep_time = self.max_timeout / 1000
        while True:
            end_time = time.time()
            if (end_time - start_time) >= self.max_timeout:
                self.msg = "Max timeout of {0} sec has reached for the execution id '{1}'.".format(self.max_timeout, task_id) + \
                           "Exiting the loop due to unexpected API '{0}' status.".format(api_name)
                self.status = "failed"
                break

            task_details = self.get_task_details(task_id)
            self.log('Getting task details from task ID {0}: {1}'.format(task_id, task_details), "DEBUG")
            if task_details.get("isError") is True:
                failure_reason = task_details.get("failureReason")
                if failure_reason:
                    self.msg = str(failure_reason)
                else:
                    self.msg = str(task_details.get("progress"))
                self.status = "failed"
                break

            for validation_string in validation_string_set:
                if validation_string in task_details.get("progress").lower():
                    self.result['changed'] = True
                    self.validation_string = validation_string
                    self.status = "success"
                    break

            if self.result['changed'] is True:
                self.log("The task with task id '{0}' is successfully executed".format(task_id), "DEBUG")
                break

            # sleep time after checking the status of the response from the API
            self.log("The time interval before checking the next response, sleep for {0}".format(sleep_time))
            time.sleep(sleep_time)
            self.log("Progress set to {0} for taskid: {1}".format(task_details.get('progress'), task_id), "DEBUG")

        return self

    def check_ise_server_updation_status(self, have_auth_details, want_auth_details):
        """
        Check if the Cisco ISE server requires an update by comparing the user name,
        FQDN, and the state of the Cisco ISE server.

        Parameters:
            have_auth_details (dict) - Current Cisco Catalyst Center authentication server information.
            want_auth_details (dict) - Desired authentication server configuration from the playbook.
        Returns:
            True or False (bool): True if the Cisco ISE server requires an update; otherwise, False.
        Description:
            Compares the user name and FQDN between the existing configuration and the new configuration.
            If there is a discrepancy, or if the server state is not 'ACTIVE', an update is required.
        """

        ip_address = have_auth_details.get("ipAddress")
        have_cisco_ise_dtos = have_auth_details.get("ciscoIseDtos")[0]
        want_cisco_ise_dtos = want_auth_details.get("ciscoIseDtos")[0]
        self.log(
            "Checking if the Cisco ISE server '{ip_address}' requires an update."
            .format(ip_address=ip_address), "DEBUG"
        )
        check_list = ["userName", "fqdn"]
        for item in check_list:
            if have_cisco_ise_dtos[item] != want_cisco_ise_dtos[item]:
                self.log(
                    "Cisco ISE server '{ip_address}' requires an update: {item} has changed."
                    .format(item=item, ip_address=ip_address), "INFO"
                )
                return True

        state = have_auth_details.get("state")
        if state != "ACTIVE":
            self.log(
                "Cisco ISE server '{ip_address}' is not in 'ACTIVE' state (current state: '{state}'). "
                "Update required.".format(ip_address=ip_address, state=state), "DEBUG"
            )
            return True

        self.log(
            "No updates are required for the Cisco ISE server '{ip_address}' based on username, fqdn, and state."
            .format(ip_address=ip_address), "DEBUG"
        )
        return False

    def format_payload_for_update(self, have_auth_server, want_auth_server):
        """
        Format the parameter of the payload for updating the authentication and policy server
        in accordance with the information in the Cisco Catalyst Ceter.

        Parameters:
            have_auth_server (dict) - Authentication and policy server information from the Cisco Catalyst Center.
            want_auth_server (dict) - Authentication and policy server information from the Playbook.

        Returns:
            self - The current object with updated desired Authentication Policy Server information.
        """

        update_params = ["authenticationPort", "accountingPort", "role"]
        for item in update_params:
            have_auth_server_item = have_auth_server.get(item)
            want_auth_server_item = want_auth_server.get(item)
            if want_auth_server_item is None:
                want_auth_server.update({item: have_auth_server_item})

            elif have_auth_server_item != want_auth_server_item:
                self.msg = "Update does not support modifying '{0}'. Here you are trying to update '{1}'." \
                           .format(update_params, item)
                self.status = "failed"
                return self

        have_auth_server_protocol = have_auth_server.get("protocol")
        want_auth_server_protocol = want_auth_server.get("protocol")
        if have_auth_server_protocol != want_auth_server_protocol:
            if want_auth_server_protocol != "RADIUS_TACACS":
                self.msg = "'protocol' can only be updated to 'RADIUS_TACACS' not from '{0}' to '{1}'" \
                           .format(have_auth_server_protocol, want_auth_server_protocol)
                self.status = "failed"
                return self

        self.log("Successfully formatted the parameter of the payload for updating the authentication and policy server.")
        self.msg = "Successfully formatted the parameter of the payload for updating the authentication and policy server."
        self.status = "success"
        return self

    def update_auth_policy_server(self, authentication_policy_server):
        """
        Update/Create Authentication and Policy Server in Cisco
        Catalyst Center with fields provided in playbook.

        Parameters:
            authentication_policy_server (List of Dict) - Playbook details of the Authentication and Policy Server.

        Returns:
            None
        """

        result_auth_server = self.result.get("response")[0].get("authenticationPolicyServer")
        auth_server_index = -1
        for item in authentication_policy_server:
            auth_server_index += 1
            ip_address = item.get("server_ip_address")
            result_auth_server.get("response").update({ip_address: {}})

            # Check Authentication and Policy Server exist, if not create and return
            is_ise_server = self.want.get("authenticationPolicyServer")[auth_server_index].get("isIseEnabled")
            if not self.have.get("authenticationPolicyServer")[auth_server_index].get("exists"):
                auth_server_params = self.want.get("authenticationPolicyServer")[auth_server_index]
                self.log("Desired State for Authentication and Policy Server for the IP '{0}' (want): {1}"
                         .format(ip_address, auth_server_params), "DEBUG")
                function_name = "add_authentication_and_policy_server_access_configuration"
                response = self.dnac._exec(
                    family="system_settings",
                    function=function_name,
                    params=auth_server_params,
                )
                if self.get_ccc_version_as_integer() >= self.get_ccc_version_as_int_from_str("2.3.7.9"):
                    validation_string_set = ("successfully created aaa settings", "operation successful")
                else:
                    validation_string_set = ("successfully created aaa settings", "operation sucessful")

                self.check_auth_server_response_status(response, validation_string_set, function_name)
                if self.status == "failed":
                    self.log(self.msg, "ERROR")
                    return

                if is_ise_server:
                    trusted_server = self.want.get("trusted_server")
                    self.check_ise_server_integration_status(ip_address)
                    self.accept_cisco_ise_server_certificate(ip_address, trusted_server)
                    ise_integration_wait_time = self.want.get("ise_integration_wait_time")
                    time.sleep(ise_integration_wait_time)
                    response = self.dnac._exec(
                        family="system_settings",
                        function='get_authentication_and_policy_servers',
                        params={"is_ise_enabled": True}
                    )
                    response = response.get("response")
                    if response is None:
                        self.msg = "Failed to retrieve the information from the API 'get_authentication_and_policy_servers' of {0}." \
                                   .format(ip_address)
                        self.status = "failed"
                        return

                    ise_server_details = get_dict_result(response, "ipAddress", ip_address)
                    ise_state_set = {"FAILED", "INPROGRESS"}
                    state = ise_server_details.get("state")
                    if state in ise_state_set:
                        if state == "INPROGRESS":
                            self.msg = "The Cisco ISE server '{ip}' integration is incomplete, currently in 'INPROGRESS' state. ".format(ip=ip_address) + \
                                       "The integration has exceeded the expected duration of '{wait_time}' second(s)." \
                                       .format(wait_time=ise_integration_wait_time)
                        elif state == "FAILED":
                            self.msg = "The Cisco ISE server '{ip}' integration has failed and in 'FAILED' state." \
                                       .format(ip=ip_address)
                            if self.want.get("trusted_server") is False:
                                self.msg += " This is the first time Cisco Catalyst Center has encountered " + \
                                            "this certificate from Cisco ISE, and it is not yet trusted."

                        self.log(str(self.msg), "ERROR")
                        self.status = "failed"
                        return

                self.log("Successfully created Authentication and Policy Server '{0}'."
                         .format(ip_address), "INFO")
                result_auth_server.get("response").get(ip_address) \
                    .update({
                        "authenticationPolicyServer Details": self.want
                            .get("authenticationPolicyServer")[auth_server_index]
                            })
                result_auth_server.get("msg").update({
                    ip_address: "Authentication and Policy Server Created Successfully"
                })
                continue

            # Authentication and Policy Server exists, check update is required
            # Edit API not working, remove this
            have_auth_server_details = self.have.get("authenticationPolicyServer")[auth_server_index].get("details")
            want_auth_server_details = self.want.get("authenticationPolicyServer")[auth_server_index]

            self.log(
                "Formatting payload for update between current and desired authentication server details.", "DEBUG"
            )
            self.format_payload_for_update(have_auth_server_details, want_auth_server_details).check_return_status()

            is_ise_server_enabled = have_auth_server_details.get("isIseEnabled")
            ise_server_requires_update = False
            if is_ise_server_enabled:
                self.log("Cisco ISE server is enabled; checking if an update is required.")
                ise_server_requires_update = self.check_ise_server_updation_status(have_auth_server_details,
                                                                                   want_auth_server_details)
                if ise_server_requires_update:
                    self.log("Cisco ISE server requires an update based on configuration changes.", "DEBUG")
                else:
                    self.log("Cisco ISE server does not require any updates.", "DEBUG")

            if not (ise_server_requires_update or self.requires_update(have_auth_server_details,
                                                                       want_auth_server_details,
                                                                       self.authentication_policy_server_obj_params)):
                self.log("Authentication and Policy Server '{0}' doesn't require an update"
                         .format(ip_address), "INFO")
                result_auth_server.get("response").get(ip_address).update({
                    "Cisco Catalyst Center params":
                    have_auth_server_details
                })
                result_auth_server.get("response").get(ip_address).update({
                    "Id": self.have.get("authenticationPolicyServer")[auth_server_index].get("id")
                })
                result_auth_server.get("msg").update({
                    ip_address: "Authentication and Policy Server doesn't require an update"
                })
                continue

            self.log("Authentication and Policy Server requires update", "DEBUG")

            # Authenticaiton and Policy Server Exists
            auth_server_params = copy.deepcopy(want_auth_server_details)
            auth_server_params.update({"id": self.have.get("authenticationPolicyServer")[auth_server_index].get("id")})
            self.log("Desired State for Authentication and Policy Server (want): {0}"
                     .format(auth_server_params), "DEBUG")
            self.log("Current State for Authentication and Policy Server (have): {0}"
                     .format(have_auth_server_details), "DEBUG")
            function_name = "edit_authentication_and_policy_server_access_configuration"
            response = self.dnac._exec(
                family="system_settings",
                function=function_name,
                params=auth_server_params,
            )
            if self.get_ccc_version_as_integer() >= self.get_ccc_version_as_int_from_str("2.3.7.9"):
                validation_string_set = ("successfully updated aaa settings", "operation successful")
            else:
                validation_string_set = ("successfully updated aaa settings", "operation sucessful")

            self.check_auth_server_response_status(response, validation_string_set, function_name)
            if self.status == "failed":
                self.log(self.msg, "ERROR")
                return

            trusted_server_msg = ""
            if is_ise_server_enabled:
                trusted_server = self.want.get("trusted_server")
                state = have_auth_server_details.get("state")
                if state != "ACTIVE":
                    self.check_ise_server_integration_status(ip_address)
                    self.accept_cisco_ise_server_certificate(ip_address, trusted_server)
                    ise_integration_wait_time = self.want.get("ise_integration_wait_time")
                    time.sleep(ise_integration_wait_time)

                ise_details = self.dnac._exec(
                    family="system_settings",
                    function='get_authentication_and_policy_servers',
                    params={"is_ise_enabled": True}
                )

                ise_details = ise_details.get("response")
                if not ise_details:
                    self.msg = (
                        "The response from the API 'get_authentication_and_policy_servers' is empty."
                    )
                    self.log(self.msg, "CRITICAL")
                    self.status = "failed"
                    return self

                if not isinstance(ise_details, list):
                    self.msg = (
                        "The response from the API 'get_authentication_and_policy_servers' "
                        "is not a dictionary."
                    )
                    self.log(self.msg, "CRITICAL")
                    self.status = "failed"
                    return self

                self.log(str(ise_details))
                state = ise_details[0].get("state")
                if not state:
                    self.msg = (
                        "The parameter 'state' is not available in the ISE details response "
                        "from the API 'get_authentication_and_policy_servers'."
                    )
                    self.status = "failed"
                    return self

                if state == "FAILED":
                    trusted_server_msg = " But the server is not trusted."

            self.log("Authentication and Policy Server '{0}' updated successfully"
                     .format(ip_address), "INFO")
            result_auth_server.get("response").get(ip_address) \
                .update({"Id": self.have.get("authenticationPolicyServer")[auth_server_index].get("id")})
            result_auth_server.get("msg").update({
                ip_address: "Authentication and Policy Server Updated Successfully.{0}".format(trusted_server_msg)
            })

        return

    def get_diff_merged(self, config):
        """
        Update or create Authentication and Policy Server in
        Cisco Catalyst Center based on the playbook details.

        Parameters:
            config (list of Dict) - Playbook details containing
            Authentication and Policy Server information.

        Returns:
            self
        """

        authentication_policy_server = config.get("authentication_policy_server")
        if authentication_policy_server is not None:
            self.update_auth_policy_server(authentication_policy_server)

        return self

    def delete_auth_policy_server(self, authentication_policy_server):
        """
        Delete a Authentication and Policy Server by server Ip address in Cisco Catalyst Center.

        Parameters:
            authentication_policy_server (List of Dict) - Playbook details of the Authentication and Policy Server.

        Returns:
            self
        """

        result_auth_server = self.result.get("response")[0].get("authenticationPolicyServer")
        auth_server_index = 0
        for item in authentication_policy_server:
            ipAddress = item.get("server_ip_address")
            auth_server_exists = self.have.get("authenticationPolicyServer")[auth_server_index].get("exists")
            if not auth_server_exists:
                result_auth_server.get("msg").update({
                    ipAddress: "Authentication and Policy Server not found."
                })
                auth_server_index += 1
                self.msg = "Authentication and Policy Server not found."
                self.status = "success"
                continue

            response = self.dnac._exec(
                family="system_settings",
                function="delete_authentication_and_policy_server_access_configuration",
                params={"id": self.have.get("authenticationPolicyServer")[auth_server_index].get("id")},
            )

            self.log("Received API response for 'delete_authentication_and_"
                     "policy_server_access_configuration': {0}".format(response), "DEBUG")

            # Check the task status
            validation_string = "successfully deleted aaa settings"
            self.check_task_response_status(response, validation_string, "delete_authentication_and_policy_server_access_configuration").check_return_status()
            taskid = response.get("response").get("taskId")

            # Update result information
            result_auth_server.get("response").update({ipAddress: {}})
            result_auth_server.get("response").get(ipAddress).update({"Task Id": taskid})
            result_auth_server.get("msg").update({
                ipAddress: "Authentication and Policy Server deleted successfully."
            })
            self.log("Authentication and Policy Server - {0} deleted successfully.".format(ipAddress))
            auth_server_index += 1

        self.msg = "Authentication and Policy Server(s) deleted successfully."
        self.status = "success"
        return self

    def get_diff_deleted(self, config):
        """
        Delete Authentication and Policy Server from the Cisco Catalyst Center.

        Parameters:
            config (list of dict) - Playbook details

        Returns:
            self
        """

        authentication_policy_server = config.get("authentication_policy_server")
        if authentication_policy_server is not None:
            self.delete_auth_policy_server(authentication_policy_server).check_return_status()

        return self

    def verify_diff_merged(self, config):
        """
        Validating the Cisco Catalyst Center configuration with the playbook details
        when state is merged (Create/Update).

        Parameters:
            config (dict) - Playbook details containing
            Authentication and Policy Server configuration.

        Returns:
            self
        """

        self.get_have(config)
        self.log("Current State (have): {0}".format(self.have), "INFO")
        self.log("Requested State (want): {0}".format(self.want), "INFO")
        if config.get("authentication_policy_server") is not None:
            self.log("Desired State of Authentication and Policy Server (want): {0}"
                     .format(self.want.get("authenticationPolicyServer")), "DEBUG")
            self.log("Current State of Authentication and Policy Server (have): {0}"
                     .format(self.have.get("authenticationPolicyServer")), "DEBUG")
            check_list = ["isIseEnabled", "ipAddress", "pxgridEnabled",
                          "useDnacCertForPxgrid", "port", "protocol",
                          "retries", "role", "timeoutSeconds", "encryptionScheme"]
            auth_server_have = self.have.get("authenticationPolicyServer")
            auth_server_want = self.want.get("authenticationPolicyServer")
            auth_server_index = 0
            total_item_in_config = len(config)
            while auth_server_index < total_item_in_config:
                for value in check_list:
                    if auth_server_have[auth_server_index].get("details").get(value) and auth_server_want[auth_server_index].get(value) and \
                            auth_server_have[auth_server_index].get("details").get(value) != auth_server_want[auth_server_index].get(value):
                        self.msg = ("Authentication and Policy Server Config of IP address '{0}' is not "
                                    "applied to the Cisco Catalyst Center.".format(config[auth_server_index].get("server_ip_address")))
                        self.status = "failed"
                        return self

                auth_server_index += 1

            self.log("Successfully validated Authentication and Policy Server(s).", "INFO")
            self.result.get("response")[0].get("authenticationPolicyServer").update({
                "Validation": "Success"
            })

        self.msg = "Successfully validated the Authentication and Policy Server."
        self.status = "success"
        return self

    def verify_diff_deleted(self, config):
        """
        Validating the Cisco Catalyst Center configuration with the playbook details
        when state is deleted (delete).

        Parameters:
            config (dict) - Playbook details containing
            Authentication and Policy Server configuration.

        Returns:
            self
        """

        self.get_have(config)
        self.log("Current State (have): {0}".format(self.have), "INFO")
        auth_server_index = 0
        authentication_policy_server = config.get("authentication_policy_server")
        for item in authentication_policy_server:
            ip_address = item.get("server_ip_address")
            self.log("Authentication and Policy Server deleted from the Cisco Catalyst Center: {0}"
                     .format(ip_address), "INFO")
            if item is not None:
                auth_server_exists = self.have.get("authenticationPolicyServer")[auth_server_index].get("exists")
                if auth_server_exists:
                    self.msg = ("Authentication and Policy Server with IP '{0}' "
                                "config is not applied to the Cisco Catalyst Center.".format(ip_address))
                    self.status = "failed"
                    return self

                self.log("Successfully validated absence of Authentication and Policy Server '{0}'."
                         .format(ip_address), "INFO")

            auth_server_index += 1

        self.result.get("response")[0].get("authenticationPolicyServer").update({
            "Validation": "Success"
        })

        self.msg = "Successfully validated the absence of Authentication and Policy Server(s)."
        self.status = "success"
        return self

    def reset_values(self):
        """
        Reset all neccessary attributes to default values

        Parameters:
            None

        Returns:
            None
        """

        self.have.clear()
        self.want.clear()
        return


def main():
    """main entry point for module execution"""

    # Define the specification for module arguments
    element_spec = {
        "dnac_host": {"type": 'str', "required": True},
        "dnac_port": {"type": 'str', "default": '443'},
        "dnac_username": {"type": 'str', "default": 'admin', "aliases": ['user']},
        "dnac_password": {"type": 'str', "no_log": True},
        "dnac_verify": {"type": 'bool', "default": 'True'},
        "dnac_version": {"type": 'str', "default": '2.2.3.3'},
        "dnac_debug": {"type": 'bool', "default": False},
        "dnac_log": {"type": 'bool', "default": False},
        "dnac_log_level": {"type": 'str', "default": 'WARNING'},
        "dnac_log_file_path": {"type": 'str', "default": 'dnac.log'},
        "dnac_log_append": {"type": 'bool', "default": True},
        "config_verify": {"type": 'bool', "default": False},
        'dnac_api_task_timeout': {'type': 'int', "default": 1200},
        'dnac_task_poll_interval': {'type': 'int', "default": 2},
        "config": {"type": 'list', "required": True, "elements": 'dict'},
        "state": {"default": 'merged', "choices": ['merged', 'deleted']},
        "validate_response_schema": {"type": 'bool', "default": True},
    }

    # Create an AnsibleModule object with argument specifications
    module = AnsibleModule(argument_spec=element_spec, supports_check_mode=False)
    ccc_ise_radius = IseRadiusIntegration(module)
    if ccc_ise_radius.compare_dnac_versions(ccc_ise_radius.get_ccc_version(), "2.3.5.3") < 0:
        ccc_ise_radius.msg = (
            "The specified version '{0}' does not support the authentication and policy server integration feature. "
            "Supported versions start from '2.3.5.3' onwards. "
            "Version '2.3.5.3' introduces APIs for creating, updating and deleting the AAA servers and ISE server."
            .format(ccc_ise_radius.get_ccc_version())
        )
        ccc_ise_radius.status = "failed"
        ccc_ise_radius.check_return_status()

    state = ccc_ise_radius.params.get("state")
    config_verify = ccc_ise_radius.params.get("config_verify")
    if state not in ccc_ise_radius.supported_states:
        ccc_ise_radius.status = "invalid"
        ccc_ise_radius.msg = "State {0} is invalid".format(state)
        ccc_ise_radius.check_return_status()

    ccc_ise_radius.validate_input().check_return_status()

    for config in ccc_ise_radius.config:
        ccc_ise_radius.reset_values()
        ccc_ise_radius.get_have(config).check_return_status()
        if state != "deleted":
            ccc_ise_radius.get_want(config).check_return_status()
        ccc_ise_radius.get_diff_state_apply[state](config).check_return_status()
        if config_verify:
            ccc_ise_radius.verify_diff_state_apply[state](config).check_return_status()

    module.exit_json(**ccc_ise_radius.result)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)


DOCUMENTATION = r"""
---
module: assurance_issues_ignore
short_description: Resource module for Assurance Issues Ignore
description:
- This module represents an alias of the module assurance_issues_ignore_v1
version_added: '6.15.0'
extends_documentation_fragment:
  - cisco.dnac.module
author: Rafael Campos (@racampos)
options:
  headers:
    description: Additional headers.
    type: dict
  issueIds:
    description: Issue Ids.
    elements: str
    type: list
requirements:
- dnacentersdk >= 2.4.9
- python >= 3.5
seealso:
- name: Cisco DNA Center documentation for Issues IgnoreTheGivenListOfIssuesV1
  description: Complete reference of the IgnoreTheGivenListOfIssuesV1 API.
  link: https://developer.cisco.com/docs/dna-center/#!ignore-the-given-list-of-issues
notes:
  - SDK Method used are
    issues.Issues.ignore_the_given_list_of_issues_v1,

  - Paths used are
    post /dna/intent/api/v1/assuranceIssues/ignore,
  - It should be noted that this module is an alias of assurance_issues_ignore_v1

"""

EXAMPLES = r"""
- name: Create
  cisco.dnac.assurance_issues_ignore:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    headers: '{{my_headers | from_json}}'
    issueIds:
    - string

"""
RETURN = r"""
dnac_response:
  description: A dictionary or list with the response returned by the Cisco DNAC Python SDK
  returned: always
  type: dict
  sample: >
    {
      "response": {
        "successfulIssueIds": [
          "string"
        ],
        "failureIssueIds": [
          "string"
        ]
      },
      "version": "string"
    }
"""



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)


DOCUMENTATION = r"""
---
module: assurance_events_query
short_description: Resource module for Assurance Events Query
description:
- This module represents an alias of the module assurance_events_query_v1
version_added: '6.15.0'
extends_documentation_fragment:
  - cisco.dnac.module
author: Rafael Campos (@racampos)
options:
  attributes:
    description: Attributes.
    elements: str
    type: list
  deviceFamily:
    description: Device Family.
    elements: str
    type: list
  endTime:
    description: End Time.
    type: int
  filters:
    description: Assurance Events Query's filters.
    elements: dict
    suboptions:
      key:
        description: Key.
        type: str
      operator:
        description: Operator.
        type: str
      value:
        description: Value.
        type: str
    type: list
  headers:
    description: Additional headers.
    type: dict
  page:
    description: Assurance Events Query's page.
    suboptions:
      limit:
        description: Limit.
        type: int
      offset:
        description: Offset.
        type: int
      sortBy:
        description: Assurance Events Query's sortBy.
        elements: dict
        suboptions:
          name:
            description: Name.
            type: str
          order:
            description: Order.
            type: str
        type: list
    type: dict
  startTime:
    description: Start Time.
    type: int
  views:
    description: Views.
    elements: str
    type: list
requirements:
- dnacentersdk >= 2.4.9
- python >= 3.5
seealso:
- name: Cisco DNA Center documentation for Devices QueryAssuranceEventsWithFiltersV1
  description: Complete reference of the QueryAssuranceEventsWithFiltersV1 API.
  link: https://developer.cisco.com/docs/dna-center/#!query-assurance-events-with-filters
notes:
  - SDK Method used are
    devices.Devices.query_assurance_events_with_filters_v1,

  - Paths used are
    post /dna/data/api/v1/assuranceEvents/query,
  - It should be noted that this module is an alias of assurance_events_query_v1

"""

EXAMPLES = r"""
- name: Create
  cisco.dnac.assurance_events_query:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    attributes:
    - string
    deviceFamily:
    - string
    endTime: 0
    filters:
    - key: string
      operator: string
      value: string
    headers: '{{my_headers | from_json}}'
    page:
      limit: 0
      offset: 0
      sortBy:
      - name: string
        order: string
    startTime: 0
    views:
    - string

"""
RETURN = r"""
dnac_response:
  description: A dictionary or list with the response returned by the Cisco DNAC Python SDK
  returned: always
  type: dict
  sample: >
    {
      "response": [
        {
          "oldRadioChannelWidth": "string",
          "clientMac": "string",
          "switchNumber": "string",
          "assocRssi": 0,
          "affectedClients": [
            "string"
          ],
          "isPrivateMac": true,
          "frequency": "string",
          "apRole": "string",
          "replacingDeviceSerialNumber": "string",
          "messageType": "string",
          "failureCategory": "string",
          "apSwitchName": "string",
          "apSwitchId": "string",
          "radioChannelUtilization": "string",
          "mnemonic": "string",
          "radioChannelSlot": 0,
          "details": "string",
          "id": "string",
          "lastApDisconnectReason": "string",
          "networkDeviceName": "string",
          "identifier": "string",
          "reasonDescription": "string",
          "vlanId": "string",
          "udnId": "string",
          "auditSessionId": "string",
          "apMac": "string",
          "deviceFamily": "string",
          "radioNoise": "string",
          "wlcName": "string",
          "apRadioOperationState": "string",
          "name": "string",
          "failureIpAddress": "string",
          "newRadioChannelList": "string",
          "duid": "string",
          "roamType": "string",
          "candidateAPs": [
            {
              "apId": "string",
              "apName": "string",
              "apMac": "string",
              "bssid": "string",
              "rssi": 0
            }
          ],
          "replacedDeviceSerialNumber": "string",
          "oldRadioChannelList": "string",
          "ssid": "string",
          "subReasonDescription": "string",
          "wirelessClientEventEndTime": 0,
          "ipv4": "string",
          "wlcId": "string",
          "ipv6": "string",
          "missingResponseAPs": [
            {
              "apId": "string",
              "apName": "string",
              "apMac": "string",
              "bssid": "string",
              "type": "string",
              "frameType": "string"
            }
          ],
          "timestamp": 0,
          "severity": 0,
          "currentRadioPowerLevel": 0,
          "newRadioChannelWidth": "string",
          "assocSnr": 0,
          "authServerIp": "string",
          "childEvents": [
            {
              "id": "string",
              "name": "string",
              "timestamp": 0,
              "wirelessEventType": 0,
              "details": "string",
              "reasonCode": "string",
              "reasonDescription": "string",
              "subReasonCode": "string",
              "subReasonDescription": "string",
              "resultStatus": "string",
              "failureCategory": "string"
            }
          ],
          "connectedInterfaceName": "string",
          "dhcpServerIp": "string",
          "managementIpAddress": "string",
          "previousRadioPowerLevel": 0,
          "resultStatus": "string",
          "radioInterference": "string",
          "networkDeviceId": "string",
          "siteHierarchy": "string",
          "eventStatus": "string",
          "wirelessClientEventStartTime": 0,
          "siteHierarchyId": "string",
          "udnName": "string",
          "facility": "string",
          "lastApResetType": "string",
          "invalidIeAPs": [
            {
              "apId": "string",
              "apName": "string",
              "apMac": "string",
              "bssid": "string",
              "type": "string",
              "frameType": "string",
              "ies": "string"
            }
          ],
          "username": "string"
        }
      ],
      "version": "string",
      "page": {
        "limit": 0,
        "offset": 0,
        "count": 0,
        "sortBy": [
          {
            "name": "string",
            "order": "string"
          }
        ]
      }
    }
"""



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: device_administration_network_conditions_info
short_description: Information module for Device Administration Network Conditions
description:
- Get all Device Administration Network Conditions.
- Get Device Administration Network Conditions by id.
- Device Admin - Returns a list of network conditions.
- Device Admin - Returns a network condition.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.ise.module_info
author: Rafael Campos (@racampos)
options:
  id:
    description:
    - Id path parameter. Condition id.
    type: str
requirements:
- ciscoisesdk >= 2.2.3
- python >= 3.5
seealso:
- name: Cisco ISE documentation for Device Administration - Network Conditions
  description: Complete reference of the Device Administration - Network Conditions API.
  link: https://developer.cisco.com/docs/identity-services-engine/v1/#!policy-openapi
notes:
  - SDK Method used are
    device_administration_network_conditions.DeviceAdministrationNetworkConditions.get_device_admin_network_condition_by_id,
    device_administration_network_conditions.DeviceAdministrationNetworkConditions.get_device_admin_network_conditions,

  - Paths used are
    get /device-admin/network-condition,
    get /device-admin/network-condition/{id},

"""

EXAMPLES = r"""
- name: Get all Device Administration Network Conditions
  cisco.ise.device_administration_network_conditions_info:
    ise_hostname: "{{ise_hostname}}"
    ise_username: "{{ise_username}}"
    ise_password: "{{ise_password}}"
    ise_verify: "{{ise_verify}}"
  register: result

- name: Get Device Administration Network Conditions by id
  cisco.ise.device_administration_network_conditions_info:
    ise_hostname: "{{ise_hostname}}"
    ise_username: "{{ise_username}}"
    ise_password: "{{ise_password}}"
    ise_verify: "{{ise_verify}}"
    id: string
  register: result

"""

RETURN = r"""
ise_response:
  description: A dictionary or list with the response returned by the Cisco ISE Python SDK
  returned: always
  type: dict
  sample: >
    {
      "conditionType": "string",
      "description": "string",
      "id": "string",
      "link": {
        "href": "string",
        "rel": "string",
        "type": "string"
      },
      "name": "string",
      "deviceList": [
        "string"
      ],
      "cliDnisList": [
        "string"
      ],
      "ipAddrList": [
        "string"
      ],
      "macAddrList": [
        "string"
      ],
      "deviceGroupList": [
        "string"
      ]
    }
"""



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)


DOCUMENTATION = r"""
---
module: event_info
short_description: Information module for Event Info
description:
- This module represents an alias of the module event_v1_info
version_added: '3.1.0'
extends_documentation_fragment:
  - cisco.dnac.module_info
author: Rafael Campos (@racampos)
options:
  headers:
    description: Additional headers.
    type: dict
  eventId:
    description:
    - EventId query parameter. The registered EventId should be provided.
    type: str
  tags:
    description:
    - Tags query parameter. The registered Tags should be provided.
    type: str
  offset:
    description:
    - Offset query parameter. The number of Registries to offset in the resultset whose default value 0.
    type: float
  limit:
    description:
    - Limit query parameter. The number of Registries to limit in the resultset whose default value 10.
    type: float
  sortBy:
    description:
    - SortBy query parameter. SortBy field name.
    type: str
  order:
    description:
    - Order query parameter.
    type: str
requirements:
- dnacentersdk >= 2.4.9
- python >= 3.5
seealso:
- name: Cisco DNA Center documentation for Event Management GetEventsV1
  description: Complete reference of the GetEventsV1 API.
  link: https://developer.cisco.com/docs/dna-center/#!get-events
notes:
  - SDK Method used are
    event_management.EventManagement.get_events_v1,

  - Paths used are
    get /dna/intent/api/v1/events,
  - It should be noted that this module is an alias of event_v1_info

"""

EXAMPLES = r"""
- name: Get all Event Info
  cisco.dnac.event_info:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    headers: "{{my_headers | from_json}}"
    eventId: string
    tags: string
    offset: 0
    limit: 0
    sortBy: string
    order: string
  register: result

"""
RETURN = r"""
dnac_response:
  description: A dictionary or list with the response returned by the Cisco DNAC Python SDK
  returned: always
  type: list
  elements: dict
  sample: >
    [
      {
        "eventId": "string",
        "nameSpace": "string",
        "name": "string",
        "description": "string",
        "version": "string",
        "category": "string",
        "domain": "string",
        "subDomain": "string",
        "type": "string",
        "tags": [
          "string"
        ],
        "severity": 0,
        "details": {},
        "subscriptionTypes": [
          "string"
        ]
      }
    ]
"""



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: s3_object
version_added: 1.0.0
short_description: Manage objects in S3
description:
  - This module allows the user to manage the objects and directories within S3 buckets. Includes
    support for creating and deleting objects and directories, retrieving objects as files or
    strings, generating download links and copying objects that are already stored in Amazon S3.
  - S3 buckets can be created or deleted using the M(amazon.aws.s3_bucket) module.
  - Compatible with AWS, DigitalOcean, Ceph, Walrus, FakeS3 and StorageGRID.
  - When using non-AWS services, O(endpoint_url) should be specified.
options:
  bucket:
    description:
      - Bucket name.
    required: true
    type: str
  dest:
    description:
      - The destination file path when downloading an object/key when O(mode=get).
      - Ignored when O(mode) is not V(get).
    type: path
  encrypt:
    description:
      - Asks for server-side encryption of the objects when O(mode=put) or O(mode=copy).
      - Ignored when O(mode) is neither V(put) nor V(copy).
    default: true
    type: bool
  encryption_mode:
    description:
      - The encryption mode to use if O(encrypt=true).
    default: AES256
    choices:
      - AES256
      - aws:kms
    type: str
  expiry:
    description:
      - Time limit (in seconds) for the URL generated and returned by S3/Walrus when performing a
        O(mode=put) or O(mode=geturl) operation.
      - Ignored when O(mode) is neither V(put) nor V(geturl).
    default: 600
    aliases: ['expiration']
    type: int
  headers:
    description:
      - Custom headers to use when O(mode=put) as a dictionary of key value pairs.
      - Ignored when O(mode) is not V(put).
    type: dict
  marker:
    description:
      - Specifies the key to start with when using list mode. Object keys are returned in
        alphabetical order, starting with key after the marker in order.
    type: str
  max_keys:
    description:
      - Max number of results to return when O(mode=list)
      - When not set, B(all) keys will be returned.
      - Ignored when O(mode) is not V(list).
    type: int
  metadata:
    description:
      - Metadata to use when O(mode=put) or O(mode=copy) as a dictionary of key value pairs.
    type: dict
  mode:
    description:
      - Switches the module behaviour between
      - 'V(put): upload'
      - 'V(get): download'
      - 'V(geturl): return download URL'
      - 'V(getstr): download object as string'
      - 'V(list): list keys'
      - 'V(create): create bucket directories'
      - 'V(delobj): delete object'
      - 'V(copy): copy object that is already stored in another bucket'
      - Support for creating and deleting buckets was removed in release 6.0.0.
        To create and manage the bucket itself please use the M(amazon.aws.s3_bucket) module.
      - Support for V(list) has been deprecated and will be removed in a release after 2026-11-01.
        The M(amazon.aws.s3_object_info) module should be used instead of V(list).
    required: true
    choices: ['get', 'put', 'create', 'geturl', 'getstr', 'delobj', 'list', 'copy']
    type: str
  object:
    description:
      - Key name of the object inside the bucket.
      - Can be used to create "virtual directories", see examples.
      - Object key names should not include the leading C(/), see
        U(https://docs.aws.amazon.com/AmazonS3/latest/userguide/object-keys.html) for more
        information.
      - Support for passing the leading V(/) has been deprecated and will be removed
        in a release after 2025-12-01.
    type: str
  sig_v4:
    description:
      - Forces the Boto SDK to use Signature Version 4.
      - Only applies to get modes, O(mode=get), O(mode=getstr), O(mode=geturl).
    default: true
    type: bool
    version_added: 5.0.0
  permission:
    description:
      - This option lets the user set the canned permissions on the object/bucket that are created.
        The permissions that can be set are V(private), V(public-read), V(public-read-write),
        V(authenticated-read) for a bucket or V(private), V(public-read), V(public-read-write),
        V(aws-exec-read), V(authenticated-read), V(bucket-owner-read), V(bucket-owner-full-control)
        for an object. Multiple permissions can be specified as a list; although only the first one
        will be used during the initial upload of the file.
      - For a full list of permissions see the AWS documentation
        U(https://docs.aws.amazon.com/AmazonS3/latest/userguide/acl-overview.html#canned-acl).
    default: ['private']
    choices:
      - "private"
      - "public-read"
      - "public-read-write"
      - "aws-exec-read"
      - "authenticated-read"
      - "bucket-owner-read"
      - "bucket-owner-full-control"
    type: list
    elements: str
  prefix:
    description:
      - Limits the response to keys that begin with the specified prefix for list mode.
    default: ""
    type: str
  version:
    description:
      - Version ID of the object inside the bucket. Can be used to get a specific version of a file
        if versioning is enabled in the target bucket.
    type: str
  overwrite:
    description:
      - Force overwrite either locally on the filesystem or remotely with the object/key.
      - Used when O(mode=put) or O(mode=get).
      - Ignored when when O(mode) is neither V(put) nor V(get).
      - Must be a Boolean, V(always), V(never), V(different) or V(latest).
      - V(true) is the same as V(always).
      - V(false) is equal to V(never).
      - When this is set to C(different) the MD5 sum of the local file is compared with the 'ETag'
        of the object/key in S3.  The ETag may or may not be an MD5 digest of the object data. See
        the ETag response header here
        U(https://docs.aws.amazon.com/AmazonS3/latest/API/RESTCommonResponseHeaders.html).
      - When O(mode=get) and O(overwrite=latest) the last modified timestamp of local file
        is compared with the 'LastModified' of the object/key in S3.
    default: 'different'
    aliases: ['force']
    type: str
  retries:
    description:
     - On recoverable failure, how many times to retry before actually failing.
    default: 0
    type: int
    aliases: ['retry']
  dualstack:
    description:
      - Enables Amazon S3 Dual-Stack Endpoints, allowing S3 communications using both IPv4 and IPv6.
      - Support for passing O(dualstack) and O(endpoint_url) at the same time has been deprecated,
        the dualstack endpoints are automatically configured using the configured O(region).
        Support will be removed in a release after 2024-12-01.
    type: bool
    default: false
  ceph:
    description:
      - Enable API compatibility with Ceph RGW.
      - It takes into account the S3 API subset working with Ceph in order to provide the same module
        behaviour where possible.
      - Requires O(endpoint_url) if O(ceph=true).
    aliases: ['rgw']
    default: false
    type: bool
  src:
    description:
      - The source file path when performing a C(put) operation.
      - One of O(content), O(content_base64) or O(src) must be specified when O(mode=put)
        otherwise ignored.
    type: path
  content:
    description:
      - The content to C(put) into an object.
      - The parameter value will be treated as a string and converted to UTF-8 before sending it to
        S3.
      - To send binary data, use the O(content_base64) parameter instead.
      - One of O(content), O(content_base64) or O(src) must be specified when O(mode=put)
        otherwise ignored.
    version_added: "1.3.0"
    type: str
  content_base64:
    description:
      - The base64-encoded binary data to C(put) into an object.
      - Use this if you need to put raw binary data, and don't forget to encode in base64.
      - One of O(content), O(content_base64) or O(src) must be specified when O(mode=put)
        otherwise ignored.
    version_added: "1.3.0"
    type: str
  ignore_nonexistent_bucket:
    description:
      - Overrides initial bucket lookups in case bucket or IAM policies are restrictive.
      - This can be useful when a user may have the C(GetObject) permission but no other
        permissions. In which case using O(mode=get) will fail unless
        O(ignore_nonexistent_bucket=true) is specified.
    type: bool
    default: false
  encryption_kms_key_id:
    description:
      - KMS key id to use when encrypting objects using O(encryption_mode=aws:kms).
      - Ignored if O(encryption_mode) is not V(aws:kms).
    type: str
  copy_src:
    description:
    - The source details of the object to copy.
    - Required if O(mode=copy).
    type: dict
    version_added: 2.0.0
    suboptions:
      bucket:
        type: str
        description:
        - The name of the source bucket.
        required: true
      object:
        type: str
        description:
        - key name of the source object.
        - if not specified, all the objects of the O(copy_src.bucket) will be copied into the specified bucket.
        required: false
      version_id:
        type: str
        description:
        - version ID of the source object.
      prefix:
        description:
        - Copy all the keys that begin with the specified prefix.
        - Ignored if O(copy_src.object) is supplied.
        default: ""
        type: str
        version_added: 6.2.0
  validate_bucket_name:
    description:
      - Whether the bucket name should be validated to conform to AWS S3 naming rules.
      - On by default, this may be disabled for S3 backends that do not enforce these rules.
      - See the Amazon documentation for more information about bucket naming rules
        U(https://docs.aws.amazon.com/AmazonS3/latest/userguide/bucketnamingrules.html).
    type: bool
    version_added: 3.1.0
    default: True
  expected_bucket_owner:
    description:
      - The account ID of the expected bucket owner.
      - If the account ID that you provide does not match the actual owner of the bucket,
        the request fails with the HTTP status code 403 Forbidden (access denied).
    type: str
    version_added: 8.2.0
author:
  - "Lester Wade (@lwade)"
  - "Sloane Hertel (@s-hertel)"
  - "Alina Buzachis (@alinabuzachis)"
notes:
  - Support for O(tags) and O(purge_tags) was added in release 2.0.0.
  - In release 5.0.0 the O(s3_url) parameter was merged into the O(endpoint_url) parameter,
    I(s3_url) remains as an alias for O(endpoint_url).
  - For Walrus O(endpoint_url) should be set to the FQDN of the endpoint with neither scheme nor path.
  - Support for the E(S3_URL) environment variable has been
    deprecated and will be removed in a release after 2024-12-01, please use the O(endpoint_url) parameter
    or the E(AWS_URL) environment variable.
  - Support for creating and deleting buckets was removed in release 6.0.0.
extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.tags
  - amazon.aws.boto3
"""

EXAMPLES = r"""
- name: Simple PUT operation
  amazon.aws.s3_object:
    bucket: mybucket
    object: /my/desired/key.txt
    src: /usr/local/myfile.txt
    mode: put

- name: PUT operation from a rendered template
  amazon.aws.s3_object:
    bucket: mybucket
    object: /object.yaml
    content: "{{ lookup('template', 'templates/object.yaml.j2') }}"
    mode: put

- name: Simple PUT operation in Ceph RGW S3
  amazon.aws.s3_object:
    bucket: mybucket
    object: /my/desired/key.txt
    src: /usr/local/myfile.txt
    mode: put
    ceph: true
    endpoint_url: "http://localhost:8000"

- name: Simple GET operation
  amazon.aws.s3_object:
    bucket: mybucket
    object: /my/desired/key.txt
    dest: /usr/local/myfile.txt
    mode: get

- name: Get a specific version of an object.
  amazon.aws.s3_object:
    bucket: mybucket
    object: /my/desired/key.txt
    version: 48c9ee5131af7a716edc22df9772aa6f
    dest: /usr/local/myfile.txt
    mode: get

- name: PUT/upload with metadata
  amazon.aws.s3_object:
    bucket: mybucket
    object: /my/desired/key.txt
    src: /usr/local/myfile.txt
    mode: put
    metadata:
      Content-Encoding: gzip
      Cache-Control: no-cache

- name: PUT/upload with custom headers
  amazon.aws.s3_object:
    bucket: mybucket
    object: /my/desired/key.txt
    src: /usr/local/myfile.txt
    mode: put
    headers: 'x-amz-grant-full-control=emailAddress=owner@example.com'

- name: List keys simple
  amazon.aws.s3_object:
    bucket: mybucket
    mode: list

- name: List keys all options
  amazon.aws.s3_object:
    bucket: mybucket
    mode: list
    prefix: /my/desired/
    marker: /my/desired/0023.txt
    max_keys: 472

- name: GET an object but don't download if the file checksums match. New in 2.0
  amazon.aws.s3_object:
    bucket: mybucket
    object: /my/desired/key.txt
    dest: /usr/local/myfile.txt
    mode: get
    overwrite: different

- name: Delete an object from a bucket
  amazon.aws.s3_object:
    bucket: mybucket
    object: /my/desired/key.txt
    mode: delobj

- name: Copy an object already stored in another bucket
  amazon.aws.s3_object:
    bucket: mybucket
    object: /my/desired/key.txt
    mode: copy
    copy_src:
      bucket: srcbucket
      object: /source/key.txt

- name: Copy all the objects with name starting with 'ansible_'
  amazon.aws.s3_object:
    bucket: mybucket
    mode: copy
    copy_src:
      bucket: srcbucket
      prefix: 'ansible_'
"""

RETURN = r"""
msg:
  description: Message indicating the status of the operation.
  returned: always
  type: str
  sample: "PUT operation complete"
url:
  description: URL of the object.
  returned: (for put and geturl operations)
  type: str
  sample: "https://my-bucket.s3.amazonaws.com/my-key.txt?AWSAccessKeyId=<access-key>&Expires=1506888865&Signature=<signature>"
expiry:
  description: Number of seconds the presigned url is valid for.
  returned: (for geturl operation)
  type: int
  sample: 600
contents:
  description: Contents of the object as string.
  returned: (for getstr operation)
  type: str
  sample: "Hello, world!"
s3_keys:
  description: List of object keys.
  returned: (for list operation)
  type: list
  elements: str
  sample:
  - prefix1/
  - prefix1/key1
  - prefix1/key2
tags:
    description: Tags of the s3 object.
    returned: always
    type: dict
    sample: {
                "Owner": "dev001",
                "env": "test"
            }
"""

import base64
import copy
import io
import mimetypes
import os
import time
from ssl import SSLError

try:
    # Beware, S3 is a "special" case, it sometimes catches botocore exceptions and
    # re-raises them as boto3 exceptions.
    import boto3
    import botocore
except ImportError:
    pass  # Handled by AnsibleAWSModule

from ansible.module_utils.basic import to_native

from ansible_collections.amazon.aws.plugins.module_utils.botocore import is_boto3_error_code
from ansible_collections.amazon.aws.plugins.module_utils.botocore import is_boto3_error_message
from ansible_collections.amazon.aws.plugins.module_utils.modules import AnsibleAWSModule
from ansible_collections.amazon.aws.plugins.module_utils.retries import AWSRetry
from ansible_collections.amazon.aws.plugins.module_utils.s3 import HAS_MD5
from ansible_collections.amazon.aws.plugins.module_utils.s3 import calculate_etag
from ansible_collections.amazon.aws.plugins.module_utils.s3 import calculate_etag_content
from ansible_collections.amazon.aws.plugins.module_utils.s3 import list_bucket_object_keys
from ansible_collections.amazon.aws.plugins.module_utils.s3 import s3_extra_params
from ansible_collections.amazon.aws.plugins.module_utils.s3 import validate_bucket_name
from ansible_collections.amazon.aws.plugins.module_utils.tagging import ansible_dict_to_boto3_tag_list
from ansible_collections.amazon.aws.plugins.module_utils.tagging import boto3_tag_list_to_ansible_dict

IGNORE_S3_DROP_IN_EXCEPTIONS = ["XNotImplemented", "NotImplemented"]


class Sigv4Required(Exception):
    pass


class S3ObjectFailure(Exception):
    def __init__(self, message=None, original_e=None):
        super().__init__(message)
        self.original_e = original_e
        self.message = message


def key_check(module, s3, bucket, obj, version=None, validate=True):
    try:
        if version:
            s3.head_object(aws_retry=True, Bucket=bucket, Key=obj, VersionId=version)
        else:
            s3.head_object(aws_retry=True, Bucket=bucket, Key=obj)
    except is_boto3_error_code("404"):
        return False
    except is_boto3_error_code("403") as e:  # pylint: disable=duplicate-except
        if validate is True:
            module.fail_json_aws(
                e,
                msg=f"Failed while looking up object (during key check) {obj}.",
            )
    except (
        botocore.exceptions.BotoCoreError,
        botocore.exceptions.ClientError,
        boto3.exceptions.Boto3Error,
    ) as e:  # pylint: disable=duplicate-except
        raise S3ObjectFailure(f"Failed while looking up object (during key check) {obj}.", e)

    return True


def etag_compare(module, s3, bucket, obj, version=None, local_file=None, content=None):
    s3_etag = _head_object(s3, bucket, obj, version=version).get("ETag")
    if local_file is not None:
        local_etag = calculate_etag(module, local_file, s3_etag, s3, bucket, obj, version)
    else:
        local_etag = calculate_etag_content(module, content, s3_etag, s3, bucket, obj, version)
    return s3_etag == local_etag


def _head_object(s3, bucket, obj, version=None):
    try:
        if version:
            key_check = s3.head_object(aws_retry=True, Bucket=bucket, Key=obj, VersionId=version)
        else:
            key_check = s3.head_object(aws_retry=True, Bucket=bucket, Key=obj)
        if not key_check:
            return {}
        key_check.pop("ResponseMetadata")
        return key_check
    except is_boto3_error_code("404"):
        return {}


def _get_object_content(module, s3, bucket, obj, version=None):
    try:
        if version:
            contents = s3.get_object(aws_retry=True, Bucket=bucket, Key=obj, VersionId=version)["Body"].read()
        else:
            contents = s3.get_object(aws_retry=True, Bucket=bucket, Key=obj)["Body"].read()
        return contents
    except is_boto3_error_code(["404", "403"]) as e:
        # AccessDenied errors may be triggered if 1) file does not exist or 2) file exists but
        # user does not have the s3:GetObject permission.
        module.fail_json_aws(e, msg=f"Could not find the key {obj}.")
    except is_boto3_error_message("require AWS Signature Version 4"):  # pylint: disable=duplicate-except
        raise Sigv4Required()
    except is_boto3_error_code("InvalidArgument") as e:  # pylint: disable=duplicate-except
        module.fail_json_aws(e, msg=f"Could not find the key {obj}.")
    except (
        botocore.exceptions.BotoCoreError,
        botocore.exceptions.ClientError,
        boto3.exceptions.Boto3Error,
    ) as e:  # pylint: disable=duplicate-except
        raise S3ObjectFailure(f"Could not find the key {obj}.", e)


def get_s3_last_modified_timestamp(s3, bucket, obj, version=None):
    last_modified = None
    obj_info = _head_object(s3, bucket, obj, version)
    if obj_info:
        last_modified = obj_info["LastModified"].timestamp()
    return last_modified


def is_local_object_latest(s3, bucket, obj, version=None, local_file=None):
    s3_last_modified = get_s3_last_modified_timestamp(s3, bucket, obj, version)
    if not os.path.exists(local_file):
        return False
    local_last_modified = os.path.getmtime(local_file)
    return s3_last_modified <= local_last_modified


def bucket_check(module, s3, bucket, validate=True):
    try:
        s3.head_bucket(aws_retry=True, Bucket=bucket)
    except is_boto3_error_code("404") as e:
        if validate:
            raise S3ObjectFailure(
                (
                    f"Bucket '{bucket}' not found (during bucket_check).  "
                    "Support for automatically creating buckets was removed in release 6.0.0.  "
                    "The amazon.aws.s3_bucket module can be used to create buckets."
                ),
                e,
            )
    except is_boto3_error_code("403") as e:  # pylint: disable=duplicate-except
        if validate:
            raise S3ObjectFailure(
                f"Permission denied accessing bucket '{bucket}' (during bucket_check).",
                e,
            )
    except (
        botocore.exceptions.BotoCoreError,
        botocore.exceptions.ClientError,
        boto3.exceptions.Boto3Error,
    ) as e:  # pylint: disable=duplicate-except
        raise S3ObjectFailure(
            f"Failed while looking up bucket '{bucket}' (during bucket_check).",
            e,
        )


def list_keys(s3, bucket, prefix=None, marker=None, max_keys=None):
    try:
        return list_bucket_object_keys(s3, bucket, prefix=prefix, start_after=marker, max_keys=max_keys)
    except (
        botocore.exceptions.ClientError,
        botocore.exceptions.BotoCoreError,
        boto3.exceptions.Boto3Error,
    ) as e:
        raise S3ObjectFailure(f"Failed while listing the keys in the bucket {bucket}", e)


def delete_key(module, s3, bucket, obj):
    if module.check_mode:
        module.exit_json(
            msg="DELETE operation skipped - running in check mode",
            changed=True,
        )
    try:
        s3.delete_object(aws_retry=True, Bucket=bucket, Key=obj)
        module.exit_json(msg=f"Object deleted from bucket {bucket}.", changed=True)
    except (
        botocore.exceptions.ClientError,
        botocore.exceptions.BotoCoreError,
        boto3.exceptions.Boto3Error,
    ) as e:
        raise S3ObjectFailure(f"Failed while trying to delete {obj}.", e)


def put_object_acl(module, s3, bucket, obj, params=None):
    try:
        if params:
            s3.put_object(aws_retry=True, **params)
        for acl in module.params.get("permission"):
            s3.put_object_acl(aws_retry=True, ACL=acl, Bucket=bucket, Key=obj)
    except is_boto3_error_code(IGNORE_S3_DROP_IN_EXCEPTIONS):
        module.warn(
            "PutObjectAcl is not implemented by your storage provider. Set the permissions parameters to the empty list"
            " to avoid this warning"
        )
    except is_boto3_error_code("AccessControlListNotSupported"):  # pylint: disable=duplicate-except
        module.warn("PutObjectAcl operation : The bucket does not allow ACLs.")
    except (
        botocore.exceptions.BotoCoreError,
        botocore.exceptions.ClientError,
        boto3.exceptions.Boto3Error,
    ) as e:  # pylint: disable=duplicate-except
        raise S3ObjectFailure(f"Failed while creating object {obj}.", e)


def create_dirkey(module, s3, bucket, obj, encrypt, expiry):
    if module.check_mode:
        module.exit_json(msg="PUT operation skipped - running in check mode", changed=True)
    params = {"Bucket": bucket, "Key": obj, "Body": b""}
    params.update(
        get_extra_params(
            encrypt,
            module.params.get("encryption_mode"),
            module.params.get("encryption_kms_key_id"),
        )
    )
    put_object_acl(module, s3, bucket, obj, params)

    # Tags
    tags, _changed = ensure_tags(s3, module, bucket, obj)

    url = put_download_url(s3, bucket, obj, expiry)

    module.exit_json(
        msg=f"Virtual directory {obj} created in bucket {bucket}",
        url=url,
        tags=tags,
        changed=True,
    )


def path_check(path):
    if os.path.exists(path):
        return True
    else:
        return False


def guess_content_type(src):
    if src:
        content_type = mimetypes.guess_type(src)[0]
        if content_type:
            return content_type

    # S3 default content type
    return "binary/octet-stream"


def get_extra_params(
    encrypt=None,
    encryption_mode=None,
    encryption_kms_key_id=None,
    metadata=None,
):
    extra = {}
    if encrypt:
        extra["ServerSideEncryption"] = encryption_mode
    if encryption_kms_key_id and encryption_mode == "aws:kms":
        extra["SSEKMSKeyId"] = encryption_kms_key_id
    if metadata:
        extra["Metadata"] = {}
        # determine object metadata and extra arguments
        for option in metadata:
            extra_args_option = option_in_extra_args(option)
            if extra_args_option:
                extra[extra_args_option] = metadata[option]
            else:
                extra["Metadata"][option] = metadata[option]
    return extra


def option_in_extra_args(option):
    temp_option = option.replace("-", "").lower()

    allowed_extra_args = {
        "acl": "ACL",
        "cachecontrol": "CacheControl",
        "contentdisposition": "ContentDisposition",
        "contentencoding": "ContentEncoding",
        "contentlanguage": "ContentLanguage",
        "contenttype": "ContentType",
        "expires": "Expires",
        "grantfullcontrol": "GrantFullControl",
        "grantread": "GrantRead",
        "grantreadacp": "GrantReadACP",
        "grantwriteacp": "GrantWriteACP",
        "metadata": "Metadata",
        "requestpayer": "RequestPayer",
        "serversideencryption": "ServerSideEncryption",
        "storageclass": "StorageClass",
        "ssecustomeralgorithm": "SSECustomerAlgorithm",
        "ssecustomerkey": "SSECustomerKey",
        "ssecustomerkeymd5": "SSECustomerKeyMD5",
        "ssekmskeyid": "SSEKMSKeyId",
        "websiteredirectlocation": "WebsiteRedirectLocation",
    }

    if temp_option in allowed_extra_args:
        return allowed_extra_args[temp_option]


def upload_s3file(
    module,
    s3,
    bucket,
    obj,
    expiry,
    metadata,
    encrypt,
    headers,
    src=None,
    content=None,
    acl_disabled=False,
):
    if module.check_mode:
        module.exit_json(msg="PUT operation skipped - running in check mode", changed=True)
    try:
        extra = get_extra_params(
            encrypt,
            module.params.get("encryption_mode"),
            module.params.get("encryption_kms_key_id"),
            metadata,
        )
        if module.params.get("permission"):
            permissions = module.params["permission"]
            if isinstance(permissions, str):
                extra["ACL"] = permissions
            elif isinstance(permissions, list):
                extra["ACL"] = permissions[0]

        if "ContentType" not in extra:
            extra["ContentType"] = guess_content_type(src)

        if src:
            s3.upload_file(aws_retry=True, Filename=src, Bucket=bucket, Key=obj, ExtraArgs=extra)
        else:
            f = io.BytesIO(content)
            s3.upload_fileobj(aws_retry=True, Fileobj=f, Bucket=bucket, Key=obj, ExtraArgs=extra)
    except (
        botocore.exceptions.ClientError,
        botocore.exceptions.BotoCoreError,
        boto3.exceptions.Boto3Error,
    ) as e:
        raise S3ObjectFailure("Unable to complete PUT operation.", e)

    if not acl_disabled:
        put_object_acl(module, s3, bucket, obj)

    # Tags
    tags, _changed = ensure_tags(s3, module, bucket, obj)

    url = put_download_url(s3, bucket, obj, expiry)

    module.exit_json(msg="PUT operation complete", url=url, tags=tags, changed=True)


def download_s3file(module, s3, bucket, obj, dest, retries, version=None):
    if module.check_mode:
        module.exit_json(msg="GET operation skipped - running in check mode", changed=True)

    optional_kwargs = {"ExtraArgs": {"VersionId": version}} if version else {}
    for x in range(0, retries + 1):
        try:
            s3.download_file(bucket, obj, dest, aws_retry=True, **optional_kwargs)
            module.exit_json(msg="GET operation complete", changed=True)
        except (
            botocore.exceptions.ClientError,
            botocore.exceptions.BotoCoreError,
            boto3.exceptions.Boto3Error,
        ) as e:
            # actually fail on last pass through the loop.
            if x >= retries:
                raise S3ObjectFailure(f"Failed while downloading {obj}.", e)
            # otherwise, try again, this may be a transient timeout.
        except SSLError as e:  # will ClientError catch SSLError?
            # actually fail on last pass through the loop.
            if x >= retries:
                module.fail_json_aws(e, msg="s3 download failed")
            # otherwise, try again, this may be a transient timeout.


def download_s3str(module, s3, bucket, obj, version=None):
    if module.check_mode:
        module.exit_json(msg="GET operation skipped - running in check mode", changed=True)
    contents = to_native(_get_object_content(module, s3, bucket, obj, version))
    module.exit_json(msg="GET operation complete", contents=contents, changed=True)


def get_download_url(module, s3, bucket, obj, expiry, tags=None, changed=True):
    try:
        url = s3.generate_presigned_url(
            # aws_retry=True,
            ClientMethod="get_object",
            Params={"Bucket": bucket, "Key": obj},
            ExpiresIn=expiry,
        )
        module.exit_json(
            msg="Download url:",
            url=url,
            tags=tags,
            expiry=expiry,
            changed=changed,
        )
    except (
        botocore.exceptions.ClientError,
        botocore.exceptions.BotoCoreError,
        boto3.exceptions.Boto3Error,
    ) as e:
        raise S3ObjectFailure("Failed while getting download url.", e)


def put_download_url(s3, bucket, obj, expiry):
    try:
        url = s3.generate_presigned_url(
            # aws_retry=True,
            ClientMethod="put_object",
            Params={"Bucket": bucket, "Key": obj},
            ExpiresIn=expiry,
        )
    except (
        botocore.exceptions.ClientError,
        botocore.exceptions.BotoCoreError,
        boto3.exceptions.Boto3Error,
    ) as e:
        raise S3ObjectFailure("Unable to generate presigned URL", e)

    return url


def get_current_object_tags_dict(module, s3, bucket, obj, version=None):
    params = {"Bucket": bucket, "Key": obj}

    if module.params.get("expected_bucket_owner"):
        params["ExpectedBucketOwner"] = module.params["expected_bucket_owner"]
    if version:
        params["VersionId"] = version

    try:
        current_tags = s3.get_object_tagging(aws_retry=True, **params).get("TagSet")
    except is_boto3_error_code(IGNORE_S3_DROP_IN_EXCEPTIONS):
        module.warn("GetObjectTagging is not implemented by your storage provider.")
        return {}
    except is_boto3_error_code(["NoSuchTagSet", "NoSuchTagSetError"]):
        return {}
    return boto3_tag_list_to_ansible_dict(current_tags)


@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def put_object_tagging(s3, bucket, obj, tags):
    s3.put_object_tagging(
        Bucket=bucket,
        Key=obj,
        Tagging={"TagSet": ansible_dict_to_boto3_tag_list(tags)},
    )


@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def delete_object_tagging(s3, bucket, obj):
    s3.delete_object_tagging(Bucket=bucket, Key=obj)


def wait_tags_are_applied(module, s3, bucket, obj, expected_tags_dict, version=None):
    for _dummy in range(0, 12):
        try:
            current_tags_dict = get_current_object_tags_dict(module, s3, bucket, obj, version)
        except (
            botocore.exceptions.ClientError,
            botocore.exceptions.BotoCoreError,
            boto3.exceptions.Boto3Error,
        ) as e:
            raise S3ObjectFailure("Failed to get object tags.", e)

        if current_tags_dict != expected_tags_dict:
            time.sleep(5)
        else:
            return current_tags_dict

    module.fail_json(
        msg="Object tags failed to apply in the expected time.",
        requested_tags=expected_tags_dict,
        live_tags=current_tags_dict,
    )


def ensure_tags(client, module, bucket, obj):
    tags = module.params.get("tags")
    purge_tags = module.params.get("purge_tags")
    changed = False

    try:
        current_tags_dict = get_current_object_tags_dict(module, client, bucket, obj)
    except (
        botocore.exceptions.BotoCoreError,
        botocore.exceptions.ClientError,
        boto3.exceptions.Boto3Error,
    ) as e:  # pylint: disable=duplicate-except
        raise S3ObjectFailure("Failed to get object tags.", e)

    # Tags is None, we shouldn't touch anything
    if tags is None:
        return current_tags_dict, changed

    if not purge_tags:
        # Ensure existing tags that aren't updated by desired tags remain
        current_copy = current_tags_dict.copy()
        current_copy.update(tags)
        tags = current_copy

    # Nothing to change, we shouldn't touch anything
    if current_tags_dict == tags:
        return current_tags_dict, changed

    if tags:
        try:
            put_object_tagging(client, bucket, obj, tags)
        except (
            botocore.exceptions.BotoCoreError,
            botocore.exceptions.ClientError,
            boto3.exceptions.Boto3Error,
        ) as e:
            raise S3ObjectFailure("Failed to update object tags.", e)
    else:
        try:
            delete_object_tagging(client, bucket, obj)
        except (
            botocore.exceptions.BotoCoreError,
            botocore.exceptions.ClientError,
            boto3.exceptions.Boto3Error,
        ) as e:
            raise S3ObjectFailure("Failed to delete object tags.", e)

    current_tags_dict = wait_tags_are_applied(module, client, bucket, obj, tags)
    changed = True

    return current_tags_dict, changed


def get_binary_content(s3_vars):
    # the content will be uploaded as a byte string, so we must encode it first
    bincontent = None
    if s3_vars.get("content"):
        bincontent = s3_vars["content"].encode("utf-8")
    if s3_vars.get("content_base64"):
        bincontent = base64.standard_b64decode(s3_vars["content_base64"])
    return bincontent


def s3_object_do_get(module, connection, connection_v4, s3_vars):
    if module.params.get("sig_v4"):
        connection = connection_v4

    keyrtn = key_check(
        module,
        connection,
        s3_vars["bucket"],
        s3_vars["object"],
        version=s3_vars["version"],
        validate=s3_vars["validate"],
    )
    if not keyrtn:
        if s3_vars["version"]:
            module.fail_json(msg=f"Key {s3_vars['object']} with version id {s3_vars['version']} does not exist.")
        module.fail_json(msg=f"Key {s3_vars['object']} does not exist.")
    if s3_vars["dest"] and path_check(s3_vars["dest"]) and s3_vars["overwrite"] != "always":
        if s3_vars["overwrite"] == "never":
            module.exit_json(
                msg="Local object already exists and overwrite is disabled.",
                changed=False,
            )
        if s3_vars["overwrite"] == "different" and etag_compare(
            module,
            connection,
            s3_vars["bucket"],
            s3_vars["object"],
            version=s3_vars["version"],
            local_file=s3_vars["dest"],
        ):
            module.exit_json(
                msg="Local and remote object are identical, ignoring. Use overwrite=always parameter to force.",
                changed=False,
            )
        if s3_vars["overwrite"] == "latest" and is_local_object_latest(
            connection,
            s3_vars["bucket"],
            s3_vars["object"],
            version=s3_vars["version"],
            local_file=s3_vars["dest"],
        ):
            module.exit_json(
                msg="Local object is latest, ignoreing. Use overwrite=always parameter to force.",
                changed=False,
            )

    try:
        download_s3file(
            module,
            connection,
            s3_vars["bucket"],
            s3_vars["object"],
            s3_vars["dest"],
            s3_vars["retries"],
            version=s3_vars["version"],
        )
    except Sigv4Required:
        download_s3file(
            module,
            connection_v4,
            s3_vars["bucket"],
            s3_vars["obj"],
            s3_vars["dest"],
            s3_vars["retries"],
            version=s3_vars["version"],
        )

    module.exit_json(failed=False)


def s3_object_do_put(module, connection, connection_v4, s3_vars):
    # if putting an object in a bucket yet to be created, acls for the bucket and/or the object may be specified
    # these were separated into the variables bucket_acl and object_acl above

    # if encryption mode is set to aws:kms then we're forced to use s3v4, no point trying the
    # original signature.
    if module.params.get("encryption_mode") == "aws:kms":
        connection = connection_v4

    if s3_vars["src"] is not None and not path_check(s3_vars["src"]):
        module.fail_json(msg=f"Local object \"{s3_vars['src']}\" does not exist for PUT operation")

    keyrtn = key_check(
        module,
        connection,
        s3_vars["bucket"],
        s3_vars["object"],
        version=s3_vars["version"],
        validate=s3_vars["validate"],
    )

    # the content will be uploaded as a byte string, so we must encode it first
    bincontent = get_binary_content(s3_vars)

    if keyrtn and s3_vars["overwrite"] != "always":
        if s3_vars["overwrite"] == "never" or etag_compare(
            module,
            connection,
            s3_vars["bucket"],
            s3_vars["object"],
            version=s3_vars["version"],
            local_file=s3_vars["src"],
            content=bincontent,
        ):
            # Return the download URL for the existing object and ensure tags are updated
            tags, tags_update = ensure_tags(connection, module, s3_vars["bucket"], s3_vars["object"])
            get_download_url(
                module,
                connection,
                s3_vars["bucket"],
                s3_vars["object"],
                s3_vars["expiry"],
                tags,
                changed=tags_update,
            )

    # only use valid object acls for the upload_s3file function
    if not s3_vars["acl_disabled"]:
        s3_vars["permission"] = s3_vars["object_acl"]
    upload_s3file(
        module,
        connection,
        s3_vars["bucket"],
        s3_vars["object"],
        s3_vars["expiry"],
        s3_vars["metadata"],
        s3_vars["encrypt"],
        s3_vars["headers"],
        src=s3_vars["src"],
        content=bincontent,
        acl_disabled=s3_vars["acl_disabled"],
    )
    module.exit_json(failed=False)


def s3_object_do_delobj(module, connection, connection_v4, s3_vars):
    # Delete an object from a bucket, not the entire bucket
    if not s3_vars.get("object", None):
        module.fail_json(msg="object parameter is required")
    elif s3_vars["bucket"] and delete_key(module, connection, s3_vars["bucket"], s3_vars["object"]):
        module.exit_json(
            msg=f"Object deleted from bucket {s3_vars['bucket']}.",
            changed=True,
        )
    else:
        module.fail_json(msg="Bucket parameter is required.")


def s3_object_do_list(module, connection, connection_v4, s3_vars):
    # If the bucket does not exist then bail out
    keys = list_keys(
        connection,
        s3_vars["bucket"],
        s3_vars["prefix"],
        s3_vars["marker"],
        s3_vars["max_keys"],
    )

    module.exit_json(msg="LIST operation complete", s3_keys=keys)


def s3_object_do_create(module, connection, connection_v4, s3_vars):
    # if both creating a bucket and putting an object in it, acls for the bucket and/or the object may be specified
    # these were separated above into the variables bucket_acl and object_acl

    if not s3_vars["object"].endswith("/"):
        s3_vars["object"] = s3_vars["object"] + "/"

    if key_check(module, connection, s3_vars["bucket"], s3_vars["object"]):
        module.exit_json(
            msg=f"Bucket {s3_vars['bucket']} and key {s3_vars['object']} already exists.",
            changed=False,
        )
    if not s3_vars["acl_disabled"]:
        # setting valid object acls for the create_dirkey function
        s3_vars["permission"] = s3_vars["object_acl"]
    create_dirkey(
        module,
        connection,
        s3_vars["bucket"],
        s3_vars["object"],
        s3_vars["encrypt"],
        s3_vars["expiry"],
    )


def s3_object_do_geturl(module, connection, connection_v4, s3_vars):
    if module.params.get("sig_v4"):
        connection = connection_v4

    if key_check(
        module,
        connection,
        s3_vars["bucket"],
        s3_vars["object"],
        version=s3_vars["version"],
        validate=s3_vars["validate"],
    ):
        tags = get_current_object_tags_dict(
            module,
            connection,
            s3_vars["bucket"],
            s3_vars["object"],
            version=s3_vars["version"],
        )
        get_download_url(
            module,
            connection,
            s3_vars["bucket"],
            s3_vars["object"],
            s3_vars["expiry"],
            tags,
        )
    module.fail_json(msg=f"Key {s3_vars['object']} does not exist.")


def s3_object_do_getstr(module, connection, connection_v4, s3_vars):
    if module.params.get("sig_v4"):
        connection = connection_v4

    if s3_vars["bucket"] and s3_vars["object"]:
        if key_check(
            module,
            connection,
            s3_vars["bucket"],
            s3_vars["object"],
            version=s3_vars["version"],
            validate=s3_vars["validate"],
        ):
            try:
                download_s3str(
                    module,
                    connection,
                    s3_vars["bucket"],
                    s3_vars["object"],
                    version=s3_vars["version"],
                )
            except Sigv4Required:
                download_s3str(
                    module,
                    connection_v4,
                    s3_vars["bucket"],
                    s3_vars["object"],
                    version=s3_vars["version"],
                )
        elif s3_vars["version"]:
            module.fail_json(msg=f"Key {s3_vars['object']} with version id {s3_vars['version']} does not exist.")
        else:
            module.fail_json(msg=f"Key {s3_vars['object']} does not exist.")


def check_object_tags(module, connection, bucket, obj):
    tags = module.params.get("tags")
    purge_tags = module.params.get("purge_tags")
    diff = False
    if tags:
        current_tags_dict = get_current_object_tags_dict(module, connection, bucket, obj)
        if not purge_tags:
            # Ensure existing tags that aren't updated by desired tags remain
            current_tags_dict.update(tags)
        diff = current_tags_dict != tags
    return diff


def calculate_object_etag(module, s3, bucket, obj, head_etag, version=None):
    etag = head_etag
    if "-" in etag:
        # object has been created using multipart upload, compute ETag using
        # object content to ensure idempotency.
        contents = _get_object_content(module, s3, bucket, obj, version)
        # Set ETag to None, to force function to compute ETag from content
        etag = calculate_etag_content(module, contents, None, s3, bucket, obj)
    return etag


def copy_object_to_bucket(module, s3, bucket, obj, encrypt, metadata, validate, src_bucket, src_obj, versionId=None):
    try:
        params = {"Bucket": bucket, "Key": obj}
        if not key_check(module, s3, src_bucket, src_obj, version=versionId, validate=validate):
            # Key does not exist in source bucket
            module.exit_json(
                msg=f"Key {src_obj} does not exist in bucket {src_bucket}.",
                changed=False,
            )

        s_obj_info = _head_object(s3, src_bucket, src_obj, version=versionId)
        d_obj_info = _head_object(s3, bucket, obj)
        do_match = True
        diff_msg = None
        if d_obj_info:
            src_etag = calculate_object_etag(module, s3, src_bucket, src_obj, s_obj_info.get("ETag"), versionId)
            dst_etag = calculate_object_etag(module, s3, bucket, obj, d_obj_info.get("ETag"))
            if src_etag != dst_etag:
                # Source and destination objects ETag differ
                do_match = False
                diff_msg = "ETag from source and destination differ"
            if do_match and metadata and metadata != d_obj_info.get("Metadata"):
                # Metadata from module inputs differs from what has been retrieved from object header
                diff_msg = "Would have update object Metadata if not running in check mode."
                do_match = False
        else:
            # The destination object does not exists
            do_match = False
            diff_msg = "Would have copy object if not running in check mode."

        if do_match:
            # S3 objects are equals, ensure tags will not be updated
            if module.check_mode:
                changed = check_object_tags(module, s3, bucket, obj)
                result = {}
                if changed:
                    result.update({"msg": "Would have update object tags if not running in check mode."})
                return changed, result

            # Ensure tags
            tags, changed = ensure_tags(s3, module, bucket, obj)
            result = {"msg": "ETag from source and destination are the same"}
            if changed:
                result = {"msg": "tags successfully updated.", "tags": tags}
            return changed, result
        # S3 objects differ
        if module.check_mode:
            return True, {"msg": diff_msg}
        else:
            changed = True
            bucketsrc = {
                "Bucket": src_bucket,
                "Key": src_obj,
            }
            if versionId:
                bucketsrc.update({"VersionId": versionId})
            params.update({"CopySource": bucketsrc})
            params.update(
                get_extra_params(
                    encrypt,
                    module.params.get("encryption_mode"),
                    module.params.get("encryption_kms_key_id"),
                    metadata,
                )
            )
            if metadata:
                # 'MetadataDirective' Specifies whether the metadata is copied from the source object or replaced
                # with metadata that's provided in the request. The default value is 'COPY', therefore when user
                # specifies a metadata we should set it to 'REPLACE'
                params.update({"MetadataDirective": "REPLACE"})
            s3.copy_object(aws_retry=True, **params)
            put_object_acl(module, s3, bucket, obj)
            # Tags
            tags, tags_updated = ensure_tags(s3, module, bucket, obj)
            msg = f"Object copied from bucket {bucketsrc['Bucket']} to bucket {bucket}."
            return changed, {"msg": msg, "tags": tags}
    except (
        botocore.exceptions.BotoCoreError,
        botocore.exceptions.ClientError,
        boto3.exceptions.Boto3Error,
    ) as e:  # pylint: disable=duplicate-except
        raise S3ObjectFailure(
            f"Failed while copying object {obj} from bucket {module.params['copy_src'].get('Bucket')}.",
            e,
        )


def s3_object_do_copy(module, connection, connection_v4, s3_vars):
    copy_src = module.params.get("copy_src")
    if not copy_src.get("object") and s3_vars["object"]:
        module.fail_json(
            msg="A destination object was specified while trying to copy all the objects from the source bucket."
        )
    src_bucket = copy_src.get("bucket")
    if not copy_src.get("object"):
        # copy recursively object(s) from source bucket to destination bucket
        # list all the objects from the source bucket
        keys = list_keys(connection, src_bucket, copy_src.get("prefix"))
        if len(keys) == 0:
            module.exit_json(msg=f"No object found to be copied from source bucket {src_bucket}.")

        changed = False
        number_keys_updated = 0
        for key in keys:
            updated, result = copy_object_to_bucket(
                module,
                connection,
                s3_vars["bucket"],
                key,
                s3_vars["encrypt"],
                s3_vars["metadata"],
                s3_vars["validate"],
                src_bucket,
                key,
                versionId=copy_src.get("version_id"),
            )
            changed |= updated
            number_keys_updated += 1 if updated else 0

        msg = f"object(s) from buckets '{src_bucket}' and '{s3_vars['bucket']}' are the same."
        if number_keys_updated:
            msg = f"{number_keys_updated} copied into bucket '{s3_vars['bucket']}'"
        module.exit_json(changed=changed, msg=msg)
    else:
        # copy single object from source bucket into destination bucket
        changed, result = copy_object_to_bucket(
            module,
            connection,
            s3_vars["bucket"],
            s3_vars["object"],
            s3_vars["encrypt"],
            s3_vars["metadata"],
            s3_vars["validate"],
            src_bucket,
            copy_src.get("object"),
            versionId=copy_src.get("version_id"),
        )
        module.exit_json(changed=changed, **result)


def populate_params(module):
    # Copy the parameters dict, we shouldn't be directly modifying it.
    variable_dict = copy.deepcopy(module.params)

    if variable_dict["validate_bucket_name"]:
        validate_bucket_name(variable_dict["bucket"])

    if variable_dict.get("overwrite") == "different" and not HAS_MD5:
        module.fail_json(msg="overwrite=different is unavailable: ETag calculation requires MD5 support")

    if variable_dict.get("overwrite") not in [
        "always",
        "never",
        "different",
        "latest",
    ]:
        if module.boolean(variable_dict["overwrite"]):
            variable_dict["overwrite"] = "always"
        else:
            variable_dict["overwrite"] = "never"

    # Bucket deletion does not require obj.  Prevents ambiguity with delobj.
    if variable_dict["object"]:
        if variable_dict.get("mode") == "delete":
            module.fail_json(msg="Parameter object cannot be used with mode=delete")
        obj = variable_dict["object"]
        # If the object starts with / remove the leading character
        if obj.startswith("/"):
            obj = obj[1:]
            variable_dict["object"] = obj
            module.deprecate(
                "Support for passing object key names with a leading '/' has been deprecated.",
                date="2025-12-01",
                collection_name="amazon.aws",
            )

    variable_dict["validate"] = not variable_dict["ignore_nonexistent_bucket"]
    variable_dict["acl_disabled"] = False

    return variable_dict


def validate_bucket(module, s3, var_dict):
    bucket_check(module, s3, var_dict["bucket"], validate=var_dict["validate"])

    try:
        ownership_controls = s3.get_bucket_ownership_controls(aws_retry=True, Bucket=var_dict["bucket"])[
            "OwnershipControls"
        ]
        if ownership_controls.get("Rules"):
            object_ownership = ownership_controls["Rules"][0]["ObjectOwnership"]
            if object_ownership == "BucketOwnerEnforced":
                var_dict["acl_disabled"] = True
    # if bucket ownership controls are not found
    except botocore.exceptions.ClientError:
        pass

    if not var_dict["acl_disabled"]:
        var_dict["object_acl"] = list(var_dict.get("permission"))

    return var_dict


def main():
    # Beware: this module uses an action plugin (plugins/action/s3_object.py)
    # so that src parameter can be either in 'files/' lookup path on the
    # controller, *or* on the remote host that the task is executed on.

    valid_modes = ["get", "put", "create", "geturl", "getstr", "delobj", "list", "copy"]
    valid_acls = [
        "private",
        "public-read",
        "public-read-write",
        "aws-exec-read",
        "authenticated-read",
        "bucket-owner-read",
        "bucket-owner-full-control",
    ]

    argument_spec = dict(
        bucket=dict(required=True),
        dest=dict(default=None, type="path"),
        encrypt=dict(default=True, type="bool"),
        encryption_mode=dict(choices=["AES256", "aws:kms"], default="AES256"),
        expiry=dict(default=600, type="int", aliases=["expiration"]),
        headers=dict(type="dict"),
        marker=dict(),
        max_keys=dict(type="int", no_log=False),
        metadata=dict(type="dict"),
        mode=dict(choices=valid_modes, required=True),
        sig_v4=dict(default=True, type="bool"),
        object=dict(),
        permission=dict(type="list", elements="str", default=["private"], choices=valid_acls),
        version=dict(default=None),
        overwrite=dict(aliases=["force"], default="different"),
        prefix=dict(default=""),
        retries=dict(aliases=["retry"], type="int", default=0),
        dualstack=dict(default=False, type="bool"),
        ceph=dict(default=False, type="bool", aliases=["rgw"]),
        src=dict(type="path"),
        content=dict(),
        content_base64=dict(),
        ignore_nonexistent_bucket=dict(default=False, type="bool"),
        encryption_kms_key_id=dict(),
        tags=dict(type="dict", aliases=["resource_tags"]),
        purge_tags=dict(type="bool", default=True),
        copy_src=dict(
            type="dict",
            options=dict(
                bucket=dict(required=True),
                object=dict(),
                prefix=dict(default=""),
                version_id=dict(),
            ),
        ),
        validate_bucket_name=dict(type="bool", default=True),
        expected_bucket_owner=dict(type="str"),
    )

    required_if = [
        ["ceph", True, ["endpoint_url"]],
        ["mode", "put", ["object"]],
        ["mode", "put", ["content", "content_base64", "src"], True],
        ["mode", "create", ["object"]],
        ["mode", "get", ["dest", "object"]],
        ["mode", "getstr", ["object"]],
        ["mode", "geturl", ["object"]],
        ["mode", "copy", ["copy_src"]],
    ]

    module = AnsibleAWSModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=required_if,
        mutually_exclusive=[["content", "content_base64", "src"]],
    )

    if module.params.get("mode") == "list":
        module.deprecate(
            (
                "Support for 'list' mode has been deprecated and will be removed in a release after "
                "2024-11-01.  Please use the amazon.aws.s3_object_info module instead."
            ),
            date="2026-11-01",
            collection_name="amazon.aws",
        )

    endpoint_url = module.params.get("endpoint_url")
    dualstack = module.params.get("dualstack")

    if dualstack and endpoint_url:
        module.deprecate(
            (
                "Support for passing both the 'dualstack' and 'endpoint_url' parameters at the same "
                "time has been deprecated."
            ),
            date="2024-12-01",
            collection_name="amazon.aws",
        )
        if "amazonaws.com" not in endpoint_url:
            module.fail_json(msg="dualstack only applies to AWS S3")

    if module.params.get("overwrite") not in ("always", "never", "different", "latest"):
        module.deprecate(
            (
                "Support for passing values of 'overwrite' other than 'always', 'never', "
                "'different' or 'latest', has been deprecated."
            ),
            date="2024-12-01",
            collection_name="amazon.aws",
        )

    extra_params = s3_extra_params(module.params, sigv4=False)
    extra_params_v4 = s3_extra_params(module.params, sigv4=True)
    retry_decorator = AWSRetry.jittered_backoff()
    try:
        s3 = module.client("s3", retry_decorator=retry_decorator, **extra_params)
        s3_v4 = module.client("s3", retry_decorator=retry_decorator, **extra_params_v4)
    except (
        botocore.exceptions.ClientError,
        botocore.exceptions.BotoCoreError,
        boto3.exceptions.Boto3Error,
    ) as e:
        module.fail_json_aws(e, msg="Failed to connect to AWS")

    s3_object_params = populate_params(module)
    s3_object_params.update(validate_bucket(module, s3, s3_object_params))

    func_mapping = {
        "get": s3_object_do_get,
        "put": s3_object_do_put,
        "delobj": s3_object_do_delobj,
        "list": s3_object_do_list,
        "create": s3_object_do_create,
        "geturl": s3_object_do_geturl,
        "getstr": s3_object_do_getstr,
        "copy": s3_object_do_copy,
    }
    func = func_mapping[s3_object_params["mode"]]
    try:
        func(module, s3, s3_v4, s3_object_params)
    except botocore.exceptions.EndpointConnectionError as e:  # pylint: disable=duplicate-except
        module.fail_json_aws(e, msg="Invalid endpoint provided")
    except S3ObjectFailure as e:
        if e.original_e:
            module.fail_json_aws(e.original_e, e.message)
        else:
            module.fail_json(e.message)

    module.exit_json(failed=False)


if __name__ == "__main__":
    main()
